try:
    from isomut2py import plot
    from isomut2py import io
    from isomut2py import HOME
    from isomut2py import SAMTOOLS_MAX_DEPTH

    import copy as __copy
    import numpy as __np
    import pandas as __pd
    import os as __os
    import subprocess as __subprocess
    import scipy as __scipy
    from Bio import SeqIO as __SeqIO
    from Bio.Seq import Seq as __Seq
    from Bio.SeqRecord import SeqRecord as __SeqRecord

except ImportError:
    print('ImportError in isomut2py.postprocess, some postprocessing functions might not work.')


def optimize_score(mutation_dataframe, control_samples,
                   FPs_per_genome, score0=0, unique_samples=None):
    """

    Optimizes score values for different mutation types (SNV, INS, DEL) and ploidies according to the list of control samples and the desired
    level of false positives in the genome. The results are stored in the score_lim_dict attribute of the MutationDetection object. If plot = True,
    plots ROC curves for all mutations types (SNV, INS, DEL) and all ploidies.

    :param mutation_dataframe: The dataframe containing the mutations. (pandas.DataFrame)
    :param control_samples: a subset of bam_filename (list of sample names) that should be considered as control samples. Control samples are defined as samples where no unique mutations are expected to be found. (list of str)
    :param FPs_per_genome: the largest number of false positives tolerated in a control sample (int)
    :param score0: Score optimization starts with score0. If a larger score value is likely to be optimal, setting score0 to a number larger than 0 can decrease computation time. (default: 0) (float)
    :param unique_samples: list of unique samples where at least one mutation is detected (default: None) (list of str)

    :returns: a dictionary containing the optimized score values for each ploidy

    """

    steps = 50

    if unique_samples is None:
        unique_samples = sorted(
            list(set([item for s in list(mutation_dataframe['sample_name'].unique()) for item in s.split(',')])))

    if sum([1 for s in control_samples if s not in unique_samples]) > 0:
        raise ValueError('List of "control_samples" is not a subset of "unique_samples".')

    total_num_of_FPs_per_genome = FPs_per_genome

    unique_ploidies = sorted(list(mutation_dataframe['ploidy'].unique()))

    mut_types_all = ['SNV', 'INS', 'DEL']

    score_lim_dict = {'SNV': [], 'INS': [], 'DEL': []}

    control_idx = []
    treated_idx = []
    for i in range(len(unique_samples)):
        if (unique_samples[i] in control_samples):
            control_idx.append(i)
        else:
            treated_idx.append(i)

    control_idx = __np.array(control_idx)
    treated_idx = __np.array(treated_idx)

    if total_num_of_FPs_per_genome is not None:
        FPs_per_ploidy = dict()
        for m in mut_types_all:
            FPs_per_ploidy[m] = dict()
            for pl in unique_ploidies:
                totalmuts_per_ploidy = \
                    mutation_dataframe[(mutation_dataframe['ploidy'] == pl) & (mutation_dataframe['type'] == m)].shape[
                        0]
                totalmuts = mutation_dataframe[(mutation_dataframe['type'] == m)].shape[0]
                if (totalmuts == 0):
                    FPs_per_ploidy[m][pl] = total_num_of_FPs_per_genome
                else:
                    FPs_per_ploidy[m][pl] = int(
                        round((float(totalmuts_per_ploidy) / totalmuts) * total_num_of_FPs_per_genome))

    for m in range(len(mut_types_all)):
        if (len(unique_ploidies) == 1):
            fp, tp = [0 for k in range(steps)], [0 for k in range(steps)]
            fp_real, tp_real = [0 for k in range(steps)], [0 for k in range(steps)]
            for score_lim, j in zip(
                    __np.linspace(score0,
                                  mutation_dataframe[mutation_dataframe['type'] == mut_types_all[m]]['score'].max(),
                                  steps),
                    range(steps)):
                muts = []
                for s in unique_samples:
                    muts.append(mutation_dataframe[(mutation_dataframe['ploidy'] == unique_ploidies[0]) &
                                                   (mutation_dataframe['sample_name'] == s) &
                                                   (mutation_dataframe['score'] > score_lim) &
                                                   (mutation_dataframe['type'] == mut_types_all[m])].shape[0])
                muts = __np.array(muts)
                fp[j], tp[j] = 1e-6 * __np.max(muts[control_idx]), 1e-6 * __np.mean(muts[treated_idx])
                fp_real[j], tp_real[j] = __np.max(muts[control_idx]), __np.mean(muts[treated_idx])

            if (total_num_of_FPs_per_genome is not None):
                fp_real = __np.array(fp_real)
                tp_real = __np.array(tp_real)
                if (len(tp_real[fp_real <= FPs_per_ploidy[mut_types_all[m]][unique_ploidies[0]]]) > 0):
                    tps = tp_real[fp_real <= FPs_per_ploidy[mut_types_all[m]][unique_ploidies[0]]][0]
                    fps = fp_real[fp_real <= FPs_per_ploidy[mut_types_all[m]][unique_ploidies[0]]][0]
                    score_lim = \
                        __np.linspace(score0,
                                      mutation_dataframe[mutation_dataframe['type'] == mut_types_all[m]]['score'].max(),
                                      steps)[
                            fp_real <= FPs_per_ploidy[mut_types_all[m]][unique_ploidies[0]]][0]
                else:
                    score_lim = 10000
                score_lim_dict[mut_types_all[m]].append(score_lim)

        else:
            for i in range(len(unique_ploidies)):
                fp, tp = [0 for k in range(steps)], [0 for k in range(steps)]
                fp_real, tp_real = [0 for k in range(steps)], [0 for k in range(steps)]
                for score_lim, j in zip(
                        __np.linspace(score0,
                                      mutation_dataframe[mutation_dataframe['type'] == mut_types_all[m]]['score'].max(),
                                      steps),
                        range(steps)):
                    muts = []
                    for s in unique_samples:
                        muts.append(mutation_dataframe[(mutation_dataframe['ploidy'] == unique_ploidies[i]) &
                                                       (mutation_dataframe['sample_name'] == s) &
                                                       (mutation_dataframe['score'] > score_lim) &
                                                       (mutation_dataframe['type'] == mut_types_all[m])].shape[0])
                    muts = __np.array(muts)
                    fp[j], tp[j] = 1e-6 * __np.max(muts[control_idx]), 1e-6 * __np.mean(muts[treated_idx])
                    fp_real[j], tp_real[j] = __np.max(muts[control_idx]), __np.mean(muts[treated_idx])

                if (total_num_of_FPs_per_genome is not None):
                    fp_real = __np.array(fp_real)
                    tp_real = __np.array(tp_real)
                    if (len(tp_real[fp_real <= FPs_per_ploidy[mut_types_all[m]][unique_ploidies[i]]]) > 0):
                        tps = tp_real[fp_real <= FPs_per_ploidy[mut_types_all[m]][unique_ploidies[i]]][0]
                        fps = fp_real[fp_real <= FPs_per_ploidy[mut_types_all[m]][unique_ploidies[i]]][0]
                        score_lim = \
                            __np.linspace(score0, mutation_dataframe[mutation_dataframe['type'] == mut_types_all[m]][
                                'score'].max(), steps)[
                                fp_real <= FPs_per_ploidy[mut_types_all[m]][unique_ploidies[i]]][0]
                    else:
                        score_lim = 10000
                    score_lim_dict[mut_types_all[m]].append(score_lim)

    return score_lim_dict


def optimize_results(sample_names,
                     control_samples,
                     FPs_per_genome,
                     plot_roc=False,
                     plot_tuning_curve=False,
                     filtered_results_file=None,
                     output_dir=None,
                     mutations_dataframe=None):
    """

    Optimizes the list of detected mutations according to the list of control samples and desired level of false positives set by the user.
    Filtered results will be loaded to the mutations attribute of the MutationDetection object.

    :param sample_names: The list of sample names included in the analysis. (list of str)
    :param control_samples: List of sample names that should be used as control samples in the sense, that no unique mutations are expected in them. (The sample names listed here must match a subset of the sample names listed in bam_filename.) (list of str)
    :param FPs_per_genome: The total number of false positives tolerated in a control sample. (int)
    :param plot_roc: If True, ROC curves will be plotted as a visual representation of the optimization process. (default: False) (boolean)
    :param plot_tuning_curve: If True, tuning curves displaying the number of mutations found in different samples with different score filters will be plotted as a visual representation of the optimization process. (default: False) (boolean)
    :param filtered_results_file: The path to the file where filtered results should be saved. (default: [output_dir]/filtered_results.csv) (str)
    :param output_dir: the path to the directory where raw mutation tables are located (default: None) (str)
    :param mutations_dataframe: the pandas.DataFrame where mutations are located (default: None) (pandas.DataFrame)

    :returns: (score_lim_dict, filtered_results)

        - score_lim_dict: a dictionary containing the optimized score values for each ploidy separately
        - filtered_results: a pandas.DataFrame containing the filtered mutations

    """

    if (len(sample_names) < 2):
        raise ValueError('Result optimization cannot be performed on less than 2 samples.')

    if sum([1 for s in control_samples if s not in sample_names]) > 0:
        raise ValueError('List of "control_samples" is not a subset of "sample_names".')

    if filtered_results_file is None:
        if output_dir is not None:
            filtered_results_file = output_dir + '/filtered_results.csv'
        else:
            filtered_results_file = 'filtered_results.csv'

    if not isinstance(mutations_dataframe, type(None)):
        if not mutations_dataframe.__class__ == __pd.core.frame.DataFrame:
            raise ValueError('Error: "mutations_dataframe" must be a pandas DataFrame.')
        elif sorted(list(mutations_dataframe.columns)) != sorted(['sample_name', 'chr', 'pos', 'type', 'score',
                                                                  'ref', 'mut', 'cov', 'mut_freq', 'cleanliness',
                                                                  'ploidy']):
            msg = 'Error: The DataFrame supplied in argument "mutations_dataframe" does not have the required columns.'
            msg += '\n'
            msg += 'Make sure to have the following columns: sample_name, chr, pos, type, score, ' \
                   'ref, mut, cov, mut_freq, cleanliness, ploidy'
            raise ValueError(msg)
        df_somatic = mutations_dataframe[~mutations_dataframe['sample_name'].str.contains(',')]

    else:
        if (not __os.path.isfile(output_dir + '/all_SNVs.isomut2')):
            raise ValueError(
                'SNV results (' + output_dir + '/all_SNVs.isomut2) could not be found, results cannot be optimized.')
        if (not __os.path.isfile(output_dir + '/all_indels.isomut2')):
            raise ValueError(
                'Indel results (' + output_dir + '/all_indels.isomut2) could not be found, results cannot be optimized.')

        __subprocess.check_call(
            'cat ' + output_dir + '/all_SNVs.isomut2 | awk \'BEGIN{FS="\t"; OFS="\t";}{if($1 !~ /,/) print $0;}\' > ' + output_dir + '/unique_SNVs.isomut2',
            shell=True)
        __subprocess.check_call(
            'cat ' + output_dir + '/all_indels.isomut2 | awk \'BEGIN{FS="\t"; OFS="\t";}{if($1 !~ /,/) print $0;}\' > ' + output_dir + '/unique_indels.isomut2',
            shell=True)

        df_SNV = __pd.read_csv(output_dir + '/unique_SNVs.isomut2', header=0,
                               names=['sample_name', 'chr', 'pos', 'type', 'score',
                                      'ref', 'mut', 'cov', 'mut_freq', 'cleanliness', 'ploidy'],
                               sep='\t',
                               low_memory=False)
        df_indel = __pd.read_csv(output_dir + '/unique_indels.isomut2', header=0,
                                 names=['sample_name', 'chr', 'pos', 'type', 'score',
                                        'ref', 'mut', 'cov', 'mut_freq', 'cleanliness', 'ploidy'],
                                 sep='\t',
                                 low_memory=False)

        __subprocess.check_call('rm ' + output_dir + '/unique_SNVs.isomut2', shell=True)
        __subprocess.check_call('rm ' + output_dir + '/unique_indels.isomut2', shell=True)

        df_somatic = __pd.concat([df_SNV, df_indel])

    df_somatic['ploidy'] = __pd.to_numeric(df_somatic['ploidy'], errors='ignore')

    unique_ploidies = sorted(list(df_somatic['ploidy'].unique()))

    if plot_tuning_curve:
        plot.__plot_tuning_curve(control_samples=control_samples,
                                 mutation_dataframe=df_somatic,
                                 return_string=False,
                                 unique_samples=sample_names)

    score_lim_dict, f = plot.__plot_roc(mutation_dataframe=df_somatic,
                                        control_samples=control_samples,
                                        FPs_per_genome=FPs_per_genome,
                                        plot_roc=plot_roc,
                                        unique_samples=sample_names)

    if mutations_dataframe is not None:
        def filter(row):
            if row['score'] < score_lim_dict[row['type']][unique_ploidies.index(row['ploidy'])]:
                return 'FILTER'
            else:
                return 'PASS'

        mutations_dataframe_filtered = __copy.deepcopy(mutations_dataframe)
        mutations_dataframe_filtered['FILTER'] = mutations_dataframe_filtered.apply(filter, axis=1)
        return score_lim_dict, mutations_dataframe_filtered[mutations_dataframe_filtered['FILTER'] != 'FILTER'].drop(
            'FILTER', axis=1)

    else:
        io.__print_filtered_results(output_dir=output_dir,
                                    filename=filtered_results_file,
                                    unique_ploidies=unique_ploidies,
                                    score_lim_dict=score_lim_dict,
                                    control_samples=control_samples,
                                    FPs_per_genome=FPs_per_genome)

        return score_lim_dict, io.load_mutations(output_dir=output_dir, filename=filtered_results_file)


def calculate_SNV_spectrum(ref_fasta, mutations_dataframe=None,
                           sample_names=None, unique_only=True, chromosomes=None,
                           output_dir=None, mutations_filename=None):
    """

    Calculates the triplet spectrum from a dataframe of relevant mutations, using the fasta file of the reference genome.

    :param ref_fasta: the path to the reference genome fasta file (str)
    :param mutations_dataframe: Dataframe containing the list of mutations to be considered. (default: None) (pandas.DataFrame)
    :param sample_names: the list of sample names included in the analysis (default: None) (list of str)
    :param unique_only: if True, only unique mutations are considered for the spectrum (default: True) (bool)
    :param chromosomes: list of chromosomes in the analysis (default: None) (list of str)
    :param output_dir: path to the directory where mutation tables are located (default: None) (str)
    :param mutations_filename: path to the mutation table(s) (default: None) (list of str)

    :returns: A dictionary containing 96-element vectors (the counts for each mutation type) as values and sample names as keys. (str: numpy.array)

    """

    if not isinstance(mutations_dataframe, type(None)):
        if not mutations_dataframe.__class__ == __pd.core.frame.DataFrame:
            msg = 'Error: Argument "mutations_dataframe" is not a pandas DataFrame.'
            raise ValueError(msg)
        elif sorted(list(mutations_dataframe.columns)) != sorted(['sample_name', 'chr', 'pos', 'type', 'score',
                                                                  'ref', 'mut', 'cov', 'mut_freq', 'cleanliness',
                                                                  'ploidy']):
            msg = 'Error: The DataFrame supplied in argument "mutations_dataframe" does not have the required columns.'
            msg += '\n'
            msg += 'Make sure to have the following columns: sample_name, chr, pos, type, score, ' \
                   'ref, mut, cov, mut_freq, cleanliness, ploidy'
            raise ValueError(msg)
    else:
        mutations_dataframe = io.load_mutations(output_dir=output_dir,
                                                filename=mutations_filename)

    mutations_dataframe = mutations_dataframe[mutations_dataframe['type'] == 'SNV']

    all_muttypes = ["ACAA", "TGTT", "ACCA", "GGTT", "ACGA", "CGTT", "ACTA", "AGTT", "CCAA", "TGGT",
                    "CCCA", "GGGT", "CCGA", "CGGT", "CCTA", "AGGT", "GCAA", "TGCT", "GCCA", "GGCT",
                    "GCGA", "CGCT", "GCTA", "AGCT", "TCAA", "TGAT", "TCCA", "GGAT", "TCGA", "CGAT",
                    "TCTA", "AGAT", "ACAG", "TGTC", "ACCG", "GGTC", "ACGG", "CGTC", "ACTG", "AGTC",
                    "CCAG", "TGGC", "CCCG", "GGGC", "CCGG", "CGGC", "CCTG", "AGGC", "GCAG", "TGCC",
                    "GCCG", "GGCC", "GCGG", "CGCC", "GCTG", "AGCC", "TCAG", "TGAC", "TCCG", "GGAC",
                    "TCGG", "CGAC", "TCTG", "AGAC", "ACAT", "TGTA", "ACCT", "GGTA", "ACGT", "CGTA",
                    "ACTT", "AGTA", "CCAT", "TGGA", "CCCT", "GGGA", "CCGT", "CGGA", "CCTT", "AGGA",
                    "GCAT", "TGCA", "GCCT", "GGCA", "GCGT", "CGCA", "GCTT", "AGCA", "TCAT", "TGAA",
                    "TCCT", "GGAA", "TCGT", "CGAA", "TCTT", "AGAA", "ATAA", "TATT", "ATCA", "GATT",
                    "ATGA", "CATT", "ATTA", "AATT", "CTAA", "TAGT", "CTCA", "GAGT", "CTGA", "CAGT",
                    "CTTA", "AAGT", "GTAA", "TACT", "GTCA", "GACT", "GTGA", "CACT", "GTTA", "AACT",
                    "TTAA", "TAAT", "TTCA", "GAAT", "TTGA", "CAAT", "TTTA", "AAAT", "ATAC", "TATG",
                    "ATCC", "GATG", "ATGC", "CATG", "ATTC", "AATG", "CTAC", "TAGG", "CTCC", "GAGG",
                    "CTGC", "CAGG", "CTTC", "AAGG", "GTAC", "TACG", "GTCC", "GACG", "GTGC", "CACG",
                    "GTTC", "AACG", "TTAC", "TAAG", "TTCC", "GAAG", "TTGC", "CAAG", "TTTC", "AAAG",
                    "ATAG", "TATC", "ATCG", "GATC", "ATGG", "CATC", "ATTG", "AATC", "CTAG", "TAGC",
                    "CTCG", "GAGC", "CTGG", "CAGC", "CTTG", "AAGC", "GTAG", "TACC", "GTCG", "GACC",
                    "GTGG", "CACC", "GTTG", "AACC", "TTAG", "TAAC", "TTCG", "GAAC", "TTGG", "CAAC",
                    "TTTG", "AAAC"]

    fasta_records = dict()
    for record in __SeqIO.parse(ref_fasta, "fasta"):
        if not chromosomes or record.id in chromosomes:
            fasta_records[record.id] = record.seq

    if sample_names is None:
        sample_names = sorted(
            list(set([item for s in list(mutations_dataframe['sample_name'].unique()) for item in s.split(',')])))

    SNVspectra = dict()

    for sample in sample_names:
        df = mutations_dataframe[mutations_dataframe['sample_name'].str.contains(sample)]
        if unique_only:
            df = df[~df['sample_name'].str.contains(',')]

        chrom_list = list(df['chr'].unique())
        mutationtype_array = []

        for chrom in chrom_list:
            search_pos_array = list(df[df['chr'] == chrom]['pos'])
            ref_array = list(df[df['chr'] == chrom]['ref'])
            alt_array = list(df[df['chr'] == chrom]['mut'])

            for i in range(len(search_pos_array)):
                search_pos = search_pos_array[i]
                ref = ref_array[i]

                neighborhood_size = 1
                start = search_pos - neighborhood_size - len(ref)
                end = search_pos + neighborhood_size
                search_seq = str(fasta_records[chrom][start:end])

                mutationtype_array.append(search_seq + alt_array[i])

        spectrum = __np.zeros(96)
        for mutation_type in mutationtype_array:
            if mutation_type in all_muttypes:
                spectrum[int(__np.floor(all_muttypes.index(mutation_type) / 2))] += 1
        SNVspectra[sample] = spectrum

    return SNVspectra


def __get_neighborhood(fasta_records, chrom, pos, mut, ref, muttype):
    """

    Gets the reference bases in a neighborhood of a given mutation.

    :param fasta_records: List of fasta records of the reference genome, loaded with SeqIO.parse. (Bio.SeqIO.SeqRecord)
    :param chrom: Chromosome of the genomic position of the mutation. (str)
    :param pos: Position of the genomic position of the mutation. (int)
    :param mut: The mutated base(s) at the mutation. (str)
    :param ref: The reference base(s) at the mutation. (str)
    :param muttype: The mutation type ("INS" or "DEL"). (str)

    :returns: tuple (search_seq_before, mut, search_seq_after)

        - search_seq_before: the sequence before the mutation (str)
        - mut: the mutated base(s) (str)
        - search_seq_after: the sequence after the mutation (str)

    """

    if muttype == 'DEL':
        neighborhood_size = len(ref) * 10
        start = pos - neighborhood_size
        end = pos + neighborhood_size + len(ref)
        search_seq_before = str(fasta_records[chrom][start:pos])
        search_seq_after = str(fasta_records[chrom][pos + len(ref):end])
        search_seq_ref = str(fasta_records[chrom][pos:pos + len(ref)])

        return search_seq_before, search_seq_ref, search_seq_after

    elif muttype == 'INS':
        neighborhood_size = len(mut) * 10
        start = pos - neighborhood_size
        end = pos + neighborhood_size
        search_seq_before = str(fasta_records[chrom][start:pos])
        search_seq_after = str(fasta_records[chrom][pos:end])

        return search_seq_before, mut, search_seq_after


def __find_number_of_repeats(seq_before, seq, seq_after):
    """

    Finds the number of repeats before or after a mutation.

    :param seq_before: the genomic sequence before the mutation (str)
    :param seq: the genomic sequence of the mutation (str)
    :param seq_after: the sequence after the mutation (str)

    :returns: number of repeats (int)

    """
    tmp_seq = seq_after
    k = 0
    while tmp_seq[:len(seq)] == seq:
        k += 1
        tmp_seq = tmp_seq[len(seq):]
    tmp_seq = seq_before[::-1]
    seq = seq[::-1]
    while tmp_seq[:len(seq)] == seq:
        k += 1
        tmp_seq = tmp_seq[len(seq):]
    return k + 1


def __find_microhomology(seq_before, seq, seq_after):
    """

    Finds the length of microhomology before or after a mutation.

    :param seq_before: the genomic sequence before the mutation (str)
    :param seq: the genomic sequence of the mutation (str)
    :param seq_after: the sequence after the mutation (str)

    :returns: number of repeats (int)

    """
    i = 1
    number_of_bases_found = 0
    while seq[:i] == seq_after[:i]:
        number_of_bases_found += 1
        i += 1
    if number_of_bases_found == 0:
        i = 1
        while seq[::-1][:i] == seq_before[::-1][:i]:
            number_of_bases_found += 1
            i += 1
    return number_of_bases_found


def __classify_indel(seq_before, seq, seq_after, muttype):
    """

    Classify an indel into the categories described here: https://www.biorxiv.org/content/early/2018/05/15/322859

    :param seq_before: the genomic sequence before the mutation (str)
    :param seq: the genomic sequence of the mutation (str)
    :param seq_after: the sequence after the mutation (str)
    :param muttype: the type of the mutation ("INS" or "DEL") (str)

    :returns: the ID of the mutation class the indel has been classified as (int)

    """
    if len(seq) == 1:
        if muttype == 'DEL':
            base_type = 0
        elif muttype == 'INS':
            base_type = 12
        k = __find_number_of_repeats(seq_before, seq, seq_after)
        if (k >= 6):
            k = 6
        t = -42
        if seq == 'T' or seq == 'A':
            t = -1
        elif seq == 'C' or seq == 'G':
            t = 0
        if t == -42:
            return -1
        mut_class = base_type + k * 2 + t
        return mut_class
    if len(seq) > 1:
        k = __find_number_of_repeats(seq_before, seq, seq_after)
        if muttype == 'DEL':
            if k == 1:
                m = __find_microhomology(seq_before, seq, seq_after)
            if k > 1 or m == 0:
                if k >= 6:
                    k = 6
                base_type = 24
                del_size = len(seq)
                if del_size >= 5:
                    del_size = 5
                mut_class = base_type + (del_size - 2) * 6 + k
                return mut_class

            elif k == 1 and m > 0:  # no repeat, but found microhomology
                if len(seq) == 2 and m == 1: mut_class = 73
                if len(seq) == 3 and m == 1: mut_class = 74
                if len(seq) == 3 and m == 2: mut_class = 75
                if len(seq) == 4 and m == 1: mut_class = 76
                if len(seq) == 4 and m == 2: mut_class = 77
                if len(seq) == 4 and m == 3: mut_class = 78
                if len(seq) >= 5 and m == 1: mut_class = 79
                if len(seq) >= 5 and m == 2: mut_class = 80
                if len(seq) >= 5 and m == 3: mut_class = 81
                if len(seq) >= 5 and m == 4: mut_class = 82
                if len(seq) >= 5 and m >= 5: mut_class = 83
                return mut_class
        elif muttype == 'INS':
            base_type = 48
            if k >= 6:
                k = 6
            ins_size = len(seq)
            if ins_size >= 5:
                ins_size = 5
            mut_class = base_type + (ins_size - 2) * 6 + k
            return mut_class


def calculate_indel_spectrum(ref_fasta, mutations_dataframe=None,
                             sample_names=None, unique_only=True, chromosomes=None,
                             output_dir=None, mutations_filename=None):

    """

    Calculates the indel spectrum from a dataframe of relevant mutations, using the fasta file of the reference genome.

    :param ref_fasta: the path to the reference genome fasta file (str)
    :param mutations_dataframe: Dataframe containing the list of mutations to be considered. (default: None) (pandas.DataFrame)
    :param sample_names: the list of sample names included in the analysis (default: None) (list of str)
    :param unique_only: if True, only unique mutations are considered for the spectrum (default: True) (bool)
    :param chromosomes: list of chromosomes in the analysis (default: None) (list of str)
    :param output_dir: path to the directory where mutation tables are located (default: None) (str)
    :param mutations_filename: path to the mutation table(s) (default: None) (list of str)

    :returns: A dictionary containing 83-element vectors (the counts for each mutation type) as values and sample names as keys. (str: numpy.array)

    """

    if not isinstance(mutations_dataframe, type(None)):
        if not mutations_dataframe.__class__ == __pd.core.frame.DataFrame:
            msg = 'Error: Argument "mutations_dataframe" is not a pandas DataFrame.'
            raise ValueError(msg)
        elif sorted(list(mutations_dataframe.columns)) != sorted(['sample_name', 'chr', 'pos', 'type', 'score',
                                                                  'ref', 'mut', 'cov', 'mut_freq', 'cleanliness',
                                                                  'ploidy']):
            msg = 'Error: The DataFrame supplied in argument "mutations_dataframe" does not have the required columns.'
            msg += '\n'
            msg += 'Make sure to have the following columns: sample_name, chr, pos, type, score, ' \
                   'ref, mut, cov, mut_freq, cleanliness, ploidy'
            raise ValueError(msg)
    else:
        mutations_dataframe = io.load_mutations(output_dir=output_dir,
                                                filename=mutations_filename)

    mutations_dataframe = mutations_dataframe[mutations_dataframe['type'].isin(['INS', 'DEL'])]

    fasta_records = dict()
    for record in __SeqIO.parse(ref_fasta, "fasta"):
        if not chromosomes or record.id in chromosomes:
            fasta_records[record.id] = record.seq

    if sample_names is None:
        sample_names = sorted(
            list(set([item for s in list(mutations_dataframe['sample_name'].unique()) for item in s.split(',')])))

    IDspectra = dict()

    for sample in sample_names:
        df = mutations_dataframe[mutations_dataframe['sample_name'].str.contains(sample)]
        if unique_only:
            df = df[~df['sample_name'].str.contains(',')]

        if df.shape[0] == 0:
            print('No indels were found for sample ' + sample)
            IDspectra[sample] = [0] * 83
        else:
            indel_spectrum = __np.zeros(83)
            for rID, r in df.iterrows():
                seq_before, seq, seq_after = __get_neighborhood(fasta_records, r['chr'], r['pos'], r['mut'],
                                                                r['ref'],
                                                                muttype=r['type'])
                idx = __classify_indel(seq_before, seq, seq_after, muttype=r['type'])
                if idx > 0:
                    indel_spectrum[idx - 1] += 1
            IDspectra[sample] = indel_spectrum

    return IDspectra


def calculate_DNV_matrix(mutations_dataframe=None,
                         sample_names=None, unique_only=True,
                         output_dir=None, mutations_filename=None):

    """

    Calculates the DNV matrix from a dataframe of relevant mutations.

    :param mutations_dataframe: Dataframe containing the list of mutations to be considered. (default: None) (pandas.DataFrame)
    :param sample_names: the list of sample names included in the analysis (default: None) (list of str)
    :param unique_only: if True, only unique mutations are considered for the spectrum (default: True) (bool)
    :param output_dir: path to the directory where mutation tables are located (default: None) (str)
    :param mutations_filename: path to the mutation table(s) (default: None) (list of str)

    :returns: A dictionary containing DNV matrices as values and sample names as keys. (str: numpy.array)

    """

    if not isinstance(mutations_dataframe, type(None)):
        if not mutations_dataframe.__class__ == __pd.core.frame.DataFrame:
            msg = 'Error: Argument "mutations_dataframe" is not a pandas DataFrame.'
            raise ValueError(msg)
        elif sorted(list(mutations_dataframe.columns)) != sorted(['sample_name', 'chr', 'pos', 'type', 'score',
                                                                  'ref', 'mut', 'cov', 'mut_freq', 'cleanliness',
                                                                  'ploidy']):
            msg = 'Error: The DataFrame supplied in argument "mutations_dataframe" does not have the required columns.'
            msg += '\n'
            msg += 'Make sure to have the following columns: sample_name, chr, pos, type, score, ' \
                   'ref, mut, cov, mut_freq, cleanliness, ploidy'
            raise ValueError(msg)
    else:
        mutations_dataframe = io.load_mutations(output_dir=output_dir,
                                                filename=mutations_filename)

    mutations_dataframe = mutations_dataframe[mutations_dataframe['type'] == 'SNV']

    if sample_names is None:
        sample_names = sorted(
            list(set([item for s in list(mutations_dataframe['sample_name'].unique()) for item in s.split(',')])))

    DNVmatrice = dict()

    base_changes = ['C>A', 'C>G', 'C>T', 'T>A', 'T>C', 'T>G', 'A>C', 'A>G', 'A>T', 'G>A', 'G>C', 'G>T']
    base_pairs = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}

    for sample in sample_names:
        df = mutations_dataframe[mutations_dataframe['sample_name'].str.contains(sample)]
        if unique_only:
            df = df[~df['sample_name'].str.contains(',')]

        if df.shape[0] == 0:
            print('No DNVs were found for sample ' + sample)
            DNVmatrice[sample] = __np.zeros((len(base_changes), len(base_changes)))
        else:
            df = df.sort_values(by=['chr', 'pos'])
            df['next pos'] = df['pos'].shift(-1)
            df['next ref'] = df['ref'].shift(-1)
            df['next mut'] = df['mut'].shift(-1)
            DNVs = df[df['pos'] + 1 == df['next pos']]

            matrix = __np.zeros((len(base_changes), len(base_changes)))
            ref1 = list(DNVs['ref'])
            alt1 = list(DNVs['mut'])
            ref2 = list(DNVs['next ref'])
            alt2 = list(DNVs['next mut'])
            for i in range(len(ref1)):
                firstone = ref1[i] + '>' + alt1[i]
                nextone = ref2[i] + '>' + alt2[i]
                if len(base_changes) - base_changes.index(firstone) <= base_changes.index(nextone):
                    firstone = base_pairs[ref2[i]] + '>' + base_pairs[alt2[i]]
                    nextone = base_pairs[ref1[i]] + '>' + base_pairs[alt1[i]]
                matrix[base_changes.index(nextone), base_changes.index(firstone)] += 1
            DNVmatrice[sample] = matrix

    return DNVmatrice


def calculate_DNV_spectrum(mutations_dataframe=None,
                           sample_names=None, unique_only=True,
                           output_dir=None, mutations_filename=None):
    """

    Calculates the indel spectrum from a dataframe of relevant mutations.

    :param mutations_dataframe: Dataframe containing the list of mutations to be considered. (default: None) (pandas.DataFrame)
    :param sample_names: the list of sample names included in the analysis (default: None) (list of str)
    :param unique_only: if True, only unique mutations are considered for the spectrum (default: True) (bool)
    :param output_dir: path to the directory where mutation tables are located (default: None) (str)
    :param mutations_filename: path to the mutation table(s) (default: None) (list of str)

    :returns: A dictionary containing DNV spectra arrays (the counts for each mutation type) as values and sample names as keys. (str: numpy.array)

    """

    if not isinstance(mutations_dataframe, type(None)):
        if not mutations_dataframe.__class__ == __pd.core.frame.DataFrame:
            msg = 'Error: Argument "mutations_dataframe" is not a pandas DataFrame.'
            raise ValueError(msg)
        elif sorted(list(mutations_dataframe.columns)) != sorted(['sample_name', 'chr', 'pos', 'type', 'score',
                                                                  'ref', 'mut', 'cov', 'mut_freq', 'cleanliness',
                                                                  'ploidy']):
            msg = 'Error: The DataFrame supplied in argument "mutations_dataframe" does not have the required columns.'
            msg += '\n'
            msg += 'Make sure to have the following columns: sample_name, chr, pos, type, score, ' \
                   'ref, mut, cov, mut_freq, cleanliness, ploidy'
            raise ValueError(msg)
    else:
        mutations_dataframe = io.load_mutations(output_dir=output_dir,
                                                filename=mutations_filename)

    mutations_dataframe = mutations_dataframe[mutations_dataframe['type'] == 'SNV']

    if sample_names is None:
        sample_names = sorted(
            list(set([item for s in list(mutations_dataframe['sample_name'].unique()) for item in s.split(',')])))

    DNVspectra = dict()

    base_pairs = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'}

    DBS_types = ["AC>CA", "AC>CG", "AC>CT", "AC>GA", "AC>GG", "AC>GT", "AC>TA", "AC>TG", "AC>TT", "AT>CA",
                 "AT>CC",
                 "AT>CG", "AT>GA", "AT>GC", "AT>TA", "CC>AA", "CC>AG", "CC>AT", "CC>GA", "CC>GG", "CC>GT",
                 "CC>TA",
                 "CC>TG", "CC>TT", "CG>AT", "CG>GC", "CG>GT", "CG>TA", "CG>TC", "CG>TT", "CT>AA", "CT>AC",
                 "CT>AG",
                 "CT>GA", "CT>GC", "CT>GG", "CT>TA", "CT>TC", "CT>TG", "GC>AA", "GC>AG", "GC>AT", "GC>CA",
                 "GC>CG",
                 "GC>TA", "TA>AT", "TA>CG", "TA>CT", "TA>GC", "TA>GG", "TA>GT", "TC>AA", "TC>AG", "TC>AT",
                 "TC>CA",
                 "TC>CG", "TC>CT", "TC>GA", "TC>GG", "TC>GT", "TG>AA", "TG>AC", "TG>AT", "TG>CA", "TG>CC",
                 "TG>CT",
                 "TG>GA", "TG>GC", "TG>GT", "TT>AA", "TT>AC", "TT>AG", "TT>CA", "TT>CC", "TT>CG", "TT>GA",
                 "TT>GC",
                 "TT>GG"]

    for sample in sample_names:
        df = mutations_dataframe[mutations_dataframe['sample_name'].str.contains(sample)]
        if unique_only:
            df = df[~df['sample_name'].str.contains(',')]

        if df.shape[0] == 0:
            print('No DNVs were found for sample ' + sample)
            DNVspectra[sample] = __np.zeros(len(DBS_types))
        else:
            df = df.sort_values(by=['chr', 'pos'])
            df['next pos'] = df['pos'].shift(-1)
            df['next ref'] = df['ref'].shift(-1)
            df['next mut'] = df['mut'].shift(-1)
            DNVs = df[df['pos'] + 1 == df['next pos']]

            spectrum = __np.zeros(len(DBS_types))
            ref1 = list(DNVs['ref'])
            alt1 = list(DNVs['mut'])
            ref2 = list(DNVs['next ref'])
            alt2 = list(DNVs['next mut'])
            for i in range(len(ref1)):
                firstone = ref1[i] + '>' + alt1[i]
                nextone = ref2[i] + '>' + alt2[i]
                if (firstone.split('>')[0] + nextone.split('>')[0] + '>' + firstone.split('>')[1] + nextone.split('>')[
                    1]
                        not in DBS_types):
                    firstone = base_pairs[ref2[i]] + '>' + base_pairs[alt2[i]]
                    nextone = base_pairs[ref1[i]] + '>' + base_pairs[alt1[i]]
                spectrum[DBS_types.index(
                    firstone.split('>')[0] + nextone.split('>')[0] + '>' + firstone.split('>')[1] + nextone.split('>')[
                        1])] += 1
            DNVspectra[sample] = spectrum

    return DNVspectra


def check_pileup(chrom_list, from_pos_list, to_pos_list, ref_fasta, input_dir, bam_filename,
                 output_dir='',
                 samtools_fullpath='samtools',
                 base_quality_limit=30,
                 samtools_flags=' -B -d ' + str(SAMTOOLS_MAX_DEPTH) + ' ',
                 print_original=True, filename=None):
    """

    Loads pileup information for a list of genomic regions.

    :param chrom_list: List of chromosomes for the regions. (list of str)
    :param from_pos_list: List of starting positions for the regions. (list of int)
    :param to_pos_list: List of ending positions for the regions. (list of int)
    :param ref_fasta: the path to the reference genome fasta file (str)
    :param bam_filename: the filenames for the bam files of the investigated samples (list of str)
    :param output_dir: path to the directory where temporary files should be saved (default: '') (str)
    :param samtools_fullpath: path to samtools on the computer (default: "samtools') (str)
    :param base_quality_limit: base quality limit for the pileup generation (default: 30) (int)
    :param samtools_flags: additional flags for samtools (default: ' -B -d ' + str(SAMTOOLS_MAX_DEPTH) + ' ') (str)
    :param print_original: If True, prints the original string generated with samtools mpileup as well (default: True) (bool)
    :param filename: If filename is specified, results will be writted to that path. Otherwise, results are written to [output_dir]/checkPileup_tmp.csv (default: None) (str)

    :returns: df: The processed results of the pileup, containing information on all samples in a pandas.DataFrame.

    """

    rmfile = False
    if filename is None:
        filename = output_dir + '/checkPileup_tmp.csv'
        rmfile = True

    __subprocess.call('rm ' + filename, shell=True)

    cmd = ' ' + samtools_fullpath + '  mpileup -Q ' + str(base_quality_limit) + ' ' + samtools_flags
    cmd += ' -f ' + ref_fasta

    if chrom_list.__class__ != list and from_pos_list.__class__ != list and to_pos_list.__class__ != list:
        cmd += ' -r ' + chrom_list + ':' + str(int(from_pos_list) + 1) + '-' + str(to_pos_list) + ' '
    else:
        if __os.path.isfile(output_dir + '/tmp_bedfile.bed'):
            __subprocess.call('rm ' + output_dir + '/tmp_bedfile.bed', shell=True)
        for c, pf, pt in zip(chrom_list, from_pos_list, to_pos_list):
            with open(output_dir + '/tmp_bedfile.bed', 'a') as f:
                f.write(c + '\t' + str(pf) + '\t' + str(pt) + '\n')
        cmd += ' -l ' + output_dir + '/tmp_bedfile.bed '
    for bam_file in bam_filename:
        cmd += input_dir + bam_file + ' '
    cmd += ' 2>> ' + output_dir + '/samtools.log '

    if print_original:
        pileup_string = __subprocess.check_output(cmd, shell=True).decode('utf-8')
        print(pileup_string)

    cmd += '| ' + HOME + '/C/checkPileup ' + str(base_quality_limit) + ' '
    for bam_file in bam_filename:
        cmd += bam_file + ' '
    cmd += '>> ' + filename

    __subprocess.call(cmd, shell=True)

    if not (chrom_list.__class__ != list and from_pos_list.__class__ != list and to_pos_list.__class__ != list):
        __subprocess.call('rm ' + output_dir + '/tmp_bedfile.bed', shell=True)

    df = __pd.read_csv(filename, sep='\t')
    if rmfile:
        __subprocess.call('rm ' + filename, shell=True)
    df['#chr'] = df['#chr'].apply(str)
    return df


def get_details_for_mutations(ref_fasta, input_dir, bam_filename,
                              mutations_dataframe=None,
                              output_dir=None,
                              mutations_filename=None,
                              samtools_fullpath='samtools',
                              base_quality_limit=30,
                              samtools_flags=' -B -d ' + str(SAMTOOLS_MAX_DEPTH) + ' '):
    """

    Get detailed results for the list of mutations contained in the mutations attribute of the object.

    :param ref_fasta: the path to the reference genome fasta file (str)
    :param input_dir: the path to the directory where bam files are located (str)
    :param bam_filename: the filenames for the bam files of the investigated samples (list of str)
    :param mutations_dataframe: a pandas.DataFrame where mutations are stored (default: None) (pandas.DataFrame)
    :param output_dir: path to the directory where temporary files should be saved (default: '') (str)
    :param mutations_filename: path to the file(s) where mutations are stored (default: None) (list of str)
    :param samtools_fullpath: path to samtools on the computer (default: "samtools') (str)
    :param base_quality_limit: base quality limit for the pileup generation (default: 30) (int)
    :param samtools_flags: additional flags for samtools (default: ' -B -d ' + str(SAMTOOLS_MAX_DEPTH) + ' ') (str)

    :returns: df_joined: A dataframe containing the detailed results. (pandas.DataFrame)

    """

    if not isinstance(mutations_dataframe, type(None)):
        if not mutations_dataframe.__class__ == __pd.core.frame.DataFrame:
            msg = 'Error: Argument "mutations_dataframe" is not a pandas DataFrame.'
            raise ValueError(msg)
        elif sorted(list(mutations_dataframe.columns)) != sorted(['sample_name', 'chr', 'pos', 'type', 'score',
                                                                  'ref', 'mut', 'cov', 'mut_freq', 'cleanliness',
                                                                  'ploidy']):
            msg = 'Error: The DataFrame supplied in argument "mutations_dataframe" does not have the required columns.'
            msg += '\n'
            msg += 'Make sure to have the following columns: sample_name, chr, pos, type, score, ' \
                   'ref, mut, cov, mut_freq, cleanliness, ploidy'
            raise ValueError(msg)
    else:
        mutations_dataframe = io.load_mutations(output_dir=output_dir,
                                                filename=mutations_filename)

    chrom_list = list(mutations_dataframe['chr'])
    from_pos_list = [int(k) - 1 for k in list(mutations_dataframe['pos'])]
    to_pos_list = list(mutations_dataframe['pos'])
    df = check_pileup(chrom_list=chrom_list,
                      from_pos_list=from_pos_list,
                      to_pos_list=to_pos_list,
                      ref_fasta=ref_fasta,
                      input_dir=input_dir,
                      bam_filename=bam_filename,
                      output_dir=output_dir,
                      samtools_fullpath=samtools_fullpath,
                      base_quality_limit=base_quality_limit,
                      samtools_flags=samtools_flags,
                      print_original=False)

    df.rename(columns={"#chr": "chr_x", "pos": "pos_x", "ref": "ref_x"}, inplace=True)

    df_joined = __pd.merge(mutations_dataframe, df, how='inner', left_on=['chr', 'pos', 'ref'],
                           right_on=['chr_x', 'pos_x', 'ref_x'])

    cols_to_keep = list(mutations_dataframe.columns)

    for bam_file in bam_filename:
        cols_to_keep.append('cov_' + bam_file)
        cols_to_keep.append('refFreq_' + bam_file)
        cols_to_keep.append('mutFreq_' + bam_file)
        df_joined['refFreq_' + bam_file] = df_joined.apply(__get_ref_freq, args=(bam_file,), axis=1)
        df_joined['mutFreq_' + bam_file] = df_joined.apply(__get_mut_freq, args=(bam_file,), axis=1)

    df_joined = df_joined[cols_to_keep]

    df_joined.sort_values(by=['chr', 'pos'], inplace=True)

    return df_joined


def __get_ref_freq(row, sample_name):
    """

    Get the reference base frequency in a row of the database containing detailed mpileup information.

    :param row: the row of the database
    :param sample_name: the name of the sample (str)

    :returns: the value of the reference base frequency (float)

    """
    col_name = row['ref'].upper() + 'freq_' + sample_name
    if row[col_name] >= 0:
        return row[col_name]
    else:
        return 0.0


def __get_mut_freq(row, sample_name):
    """

    Get the mutated base frequency in a row of the database containing detailed mpileup information.

    :param row: the row of the database
    :param sample_name: the name of the sample (str)

    :returns: the value of the mutated base frequency (float)

    """
    if row['type'] == 'SNV':
        col_name = row['mut'].upper() + 'freq_' + sample_name
    elif row['type'] == 'INS':
        col_name = 'insfreq_' + sample_name
    elif row['type'] == 'DEL':
        col_name = 'delfreq_' + sample_name
    if row[col_name] >= 0:
        return row[col_name]
    else:
        return 0.0


def __get_cosine_similarity(spectrum1, spectrum2):
    """

    Get cosine similarity between two vectors.

    :param spectrum1: vector1 (list, numpy.array)
    :param spectrum2: vector2 (list, numpy.array)

    :returns: cosine similarity (float)

    """
    return 1 - __scipy.spatial.distance.cosine(spectrum1, spectrum2)


def __initialize_proportions(spectrum, sig_matrix, equal=False):
    """

    Initialize the initial proportions of the signature mixture.

    :param spectrum: The spectrum to be decomposed. (list, numpy.array)
    :param sig_matrix: The matrix containing the weights of each mutational type in each reference signature. (rows: mutational types, columns: reference signatures)
    :param equal: If True, initial weights are initialized to be equal for all reference signature. Otherwise, they are initialized so the signatures with large cosine similarity with the original spectrum get larger weights. (default: False) (bool)

    :returns: The initial proportions of the signatures. (numpy.array)

    """
    cos_similarities = []
    for sig in sig_matrix.T:
        cos_similarities.append(__get_cosine_similarity(spectrum, sig))
    init_prop = __np.array(cos_similarities) / __np.sum(cos_similarities)

    if equal:
        return __np.ones(len(init_prop)) / len(init_prop)

    return init_prop


def __get_delta_jg(spectrum, sig_matrix):
    """

    Initialize the probability matrix of each mutation in the spectrum originating from the given signature.

    :param spectrum: Original spectrum to decompose. (numpy.array)
    :param sig_matrix: The matrix containing the weights of each mutational type in each reference signature. (rows: mutational types, columns: reference signatures)

    :returns: delta_matrix: A matrix containing each mutation as a row, with the probability of it originating from a specific reference signature as a column. (numpy.array)

    """
    delta_matrix = __np.zeros((int(__np.sum(spectrum)), sig_matrix.shape[1]))

    i = 0
    for sID, s in enumerate(spectrum):
        for k in range(int(s)):
            delta_matrix[i, :] = sig_matrix[sID, :]
            i += 1
    return delta_matrix


def __calculate_z_jg(theta, delta_matrix):
    """

    Calculate the conditional probability of a mutation originating from a specific signature, given the proportion of the reference signatures in
    the mixture.

    :param theta: The current proportions of the reference signatures in the mixture. (numpy.array)
    :param delta_matrix: The original probability matrix of a given mutation originating from a specific reference signature. (numpy.array)

    :returns: z_matrix: A matrix containing the conditional probabilities of a mutation (row) arising from a specific reference signature (column), given the current mixture proportions of the signatures. (numpy.array)

    """
    z_matrix = __np.divide((theta * delta_matrix).T, (theta * delta_matrix).sum(axis=1)).T
    return z_matrix


def __calculate_theta(z_matrix):
    """

    Calculate the new mixture proportions from the conditional probability matrix.

    :param z_matrix: Conditional probability matrix. (numpy.array)

    :returns: theta: Updated mixture proportions. (numpy.array)

    """
    theta = z_matrix.sum(axis=0) / z_matrix.sum()
    return theta


def __check_if_converged(theta, prev_theta, tol=0.0001):
    """

    Check if the iterations have converged, ie. the old and the updated mixture proportions do not differ much.

    :param theta: Current mixture proportions. (numpy.array)
    :param prev_theta: Previous mixture proportions. (numpy.array)
    :param tol: The maximal difference between the proportions for convergence to be True. (default: 0.0001) (float)

    :returns: True if converged, False otherwise (boolean)

    """
    if (__np.sum(__np.abs(__np.exp(prev_theta) - __np.exp(theta))) < tol):
        return True
    else:
        return False


def __reconstruct_spectrum(theta, sig_matrix):
    """

    Reconstruct the spectrum using the estimated mixture proportions of the reference signatures and the signature matrix.

    :param theta: estimated mixture proportions (numpy.array)
    :param sig_matrix: The matrix containing the weights of each mutational type in each reference signature. (rows: mutational types, columns: reference signatures) (numpy.array)

    :returns: rec_spectrum: Reconstructed spectrum. (numpy.array)

    """
    rec_spectrum = (theta * sig_matrix).sum(axis=1) / (theta * sig_matrix).sum()
    return rec_spectrum


def __EM(spectrum, sig_matrix,
         max_iter=1000,
         equal_initial_proportions=False,
         tol=0.0001):
    """

    Perform expectation maximization to infer the proportions of reference signatures in a spectrum.

    :param spectrum: The original spectrum (counts). (numpy.array)
    :param sig_matrix: The matrix containing the weights of each mutational type in each reference signature. (rows: mutational types, columns: reference signatures) (numpy.array)
    :param max_iter: The maximum number of iterations (default: 1000) (int)
    :param equal_initial_proportions: If True, initial weights are initialized to be equal for all reference signature. Otherwise, they are initialized so the signatures with large cosine similarity with the original spectrum get larger weights. (default: False) (bool)
    :param tol: The maximal difference between the proportions for convergence to be True. (default: 0.0001) (float)

    :returns: (theta, z_matrix)

        - theta: The final mixture proportions. (numpy.array)
        - z_matrix: The final set of conditional probabilities, that a mutation originated from a specific reference signature, given the final mixture proportions.

    """
    # initialize proportions
    theta_prev = __initialize_proportions(spectrum, sig_matrix, equal=equal_initial_proportions)

    # calculate d_matrix
    delta_matrix = __get_delta_jg(spectrum, sig_matrix)

    for i in range(max_iter):
        z_matrix = __calculate_z_jg(theta_prev, delta_matrix)
        theta = __calculate_theta(z_matrix)
        if (__check_if_converged(theta, theta_prev, tol=tol)):
            break
        else:
            theta_prev = theta

    return theta, z_matrix


def __filter_sigs(sig_matrix, z_matrix, filter_percent=0, filter_count=0, keep_top_n=None):
    """

    Filter signatures with negligible contribution to the mixture.

    :param sig_matrix: The matrix containing the weights of each mutational type in each reference signature. (rows: mutational types, columns: reference signatures) (numpy.array)
    :param z_matrix: Conditional probability matrix. (numpy.array)
    :param filter_percent: Filter signatures, that contribute less than filter_percent of the mutations in the sample. (default: 0) (float)
    :param filter_count: Filter signatures, that contribute less than filter_count number of mutations in the sample. (default: 0) (int)
    :param keep_top_n: Only keep those signatures that contribute to the mixture with the top keep_top_n number of mutation. (default: None) (int)

    :returns: The filtered signature matrix without the filtered signatures. (numpy.array)

    """
    sig_IDs = z_matrix.argmax(axis=1)
    unique_sigs, sig_counts = __np.unique(sig_IDs, return_counts=True)
    sig_percents = sig_counts * 100 / sig_counts.sum()

    realSig_IDs = unique_sigs[(sig_counts >= filter_count) * (sig_percents >= filter_percent)]

    if keep_top_n:
        realSig_IDs = unique_sigs[sig_counts.argsort()[-keep_top_n:][::-1]]

    return sig_matrix[:, realSig_IDs], realSig_IDs


def __read_signatures_files(fn, own_file=False, sig_type='indel'):
    """

    Read signature matrix from file.

    :param fn: The path to the csv file containing the signature matrix. (str)
    :param own_file: If using an own file, set to True. (default: False) (bool)
    :param sig_type: Signature type ('indel', 'SNV' or 'DNV') (default: "indel") (str)

    :returns: (sig_matrix, sig_names)

        - sig_matrix: The matrix containing the weights of each mutational type in each reference signature. (rows: mutational types,
        columns: reference signatures) (numpy.array)
        - sig_names: Names of the signatures (numpy.array)

    """
    df_sigs = __pd.read_csv(fn)

    sig_type_string = 'ID'
    if sig_type == 'SNV':
        sig_type_string = 'SBS'
    elif sig_type == 'DNV':
        sig_type_string = 'DBS'

    if own_file:
        sig_names = [k for k in list(df_sigs.columns)]
    else:
        sig_names = [k for k in list(df_sigs.columns) if sig_type_string in k]

    sig_matrix = __np.array(df_sigs[sig_names])

    return sig_matrix, sig_names


def __decompose_spectrum(spectrum, sig_matrix, sig_names,
                         equal_initial_proportions=False,
                         tol=0.0001,
                         max_iter=1000,
                         filter_percent=0,
                         filter_count=0,
                         keep_top_n=None,
                         use_signatures=None):
    """

    Decompose a specific spectrum to a mixture of reference signatures.

    :param spectrum: The original spectrum (counts). (numpy.array)
    :param sig_matrix: The matrix containing the weights of each mutational type in each reference signature. (rows: mutational types, columns: reference signatures) (numpy.array)
    :param sig_names: Names of the signatures (numpy.array)
    :param equal_initial_proportions: If True, initial weights are initialized to be equal for all reference signature. Otherwise, they are initialized so the signatures with large cosine similarity with the original spectrum get larger weights. (default: False) (bool)
    :param tol: The maximal difference between the proportions for convergence to be True. (default: 0.0001) (float)
    :param max_iter: The maximum number of iterations (default: 1000) (int)
    :param filter_percent: Filter signatures, that contribute less than filter_percent of the mutations in the sample. (default: 0) (float)
    :param filter_count: Filter signatures, that contribute less than filter_count number of mutations in the sample. (default: 0) (int)
    :param keep_top_n: Only keep those signatures that contribute to the mixture with the top keep_top_n number of mutation. (default: None) (int)
    :param use_signatures: Use a specific subset of all signatures. A list of signature names. (default: None) (list of str)

    :returns: The final proportions of all signatures in the mixture. (Filtered out or not used signatures appear with a proportion of zero.) (numpy.array)

    """
    if use_signatures != None:
        use_IDs = __np.array([sig_names.index(us) for us in use_signatures])
        sig_matrix = sig_matrix[:, use_IDs]
        sig_names = __np.array(sig_names)[use_IDs]

    # do one round of EM
    final_theta, final_z_matrix = __EM(spectrum, sig_matrix,
                                       max_iter=max_iter,
                                       tol=tol,
                                       equal_initial_proportions=equal_initial_proportions)

    # filter potential contributors
    new_sig_matrix, realSig_IDs = __filter_sigs(sig_matrix, final_z_matrix,
                                                filter_percent=filter_percent,
                                                filter_count=filter_count,
                                                keep_top_n=keep_top_n)

    # do next round of EM
    final_theta, final_z_matrix = __EM(spectrum, new_sig_matrix,
                                       max_iter=max_iter,
                                       tol=tol,
                                       equal_initial_proportions=equal_initial_proportions)

    all_theta = []
    k = 0
    for i in range(len(sig_names)):
        if i in realSig_IDs:
            all_theta.append(final_theta[k])
            k += 1
        else:
            all_theta.append(0)

    return all_theta


def decompose_indel_spectra(IDspectrumDict,
                            sample_names=None,
                            signatures_file=None,
                            equal_initial_proportions=False,
                            tol=0.0001,
                            max_iter=1000,
                            filter_percent=0,
                            filter_count=0,
                            keep_top_n=None,
                            use_signatures=None,
                            ignore_signatures=None):
    """

    Run the whole pipeline of decomposing indel spectra for the samples specified in sample_names.

    :param IDspectrumDict: dictionary containing indel spectra as values and sample names as keys (dictionary)
    :param sample_names: The list of sample names analysed. (list of str)
    :param signatures_file: The path to the csv file containing the signature matrix. (str)
    :param equal_initial_proportions: If True, initial weights are initialized to be equal for all reference signature. Otherwise, they are initialized so the signatures with large cosine similarity with the original spectrum get larger weights. (default: False) (bool)
    :param tol: The maximal difference between the proportions for convergence to be True. (default: 0.0001) (float)
    :param max_iter: The maximum number of iterations (default: 1000) (int)
    :param filter_percent: Filter signatures, that contribute less than filter_percent of the mutations in the sample. (default: 0) (float)
    :param filter_count: Filter signatures, that contribute less than filter_count number of mutations in the sample. (default: 0) (int)
    :param keep_top_n: Only keep those signatures that contribute to the mixture with the top keep_top_n number of mutation. (default: None) (int)
    :param use_signatures: Use a specific subset of all signatures. A list of signature names. (default: None) (list of str)
    :param ignore_signatures: Exclude a specific subset of all signatures from the analysis. A list of signature names. (default: None) (list of str)

    :returns: The final proportions of all signatures in the mixture. (Filtered out or not used signatures appear with a proportion of zero.) (numpy.array)

    """

    if sample_names == None:
        sample_names = sorted(list(IDspectrumDict.keys()))

    if signatures_file == None:
        signatures_file = HOME + '/alexandrovSignatures/sigProfiler_ID_signatures.csv'

    sig_matrix, sig_names = __read_signatures_files(signatures_file, sig_type='indel')

    sig_names_filtered = sig_names
    if use_signatures is not None:
        sig_names_filtered = sorted(use_signatures)
    if ignore_signatures is not None:
        sig_names_filtered = sorted(list(set(sig_names_filtered) - set(ignore_signatures)))

    theta_matrix = []
    for s in sample_names:
        theta = __decompose_spectrum(IDspectrumDict[s], sig_matrix, sig_names,
                                     equal_initial_proportions=equal_initial_proportions,
                                     tol=tol,
                                     max_iter=max_iter,
                                     filter_percent=filter_percent,
                                     filter_count=filter_count,
                                     keep_top_n=keep_top_n,
                                     use_signatures=sig_names_filtered)
        theta_matrix.append(theta)

    return plot.__plot_spectrum_decomposition(sample_names, theta_matrix, sig_names_filtered, spectrum_type='indel')


def decompose_DNV_spectra(DNVspectrumDict,
                          sample_names=None,
                          signatures_file=None,
                          equal_initial_proportions=False,
                          tol=0.0001,
                          max_iter=1000,
                          filter_percent=0,
                          filter_count=0,
                          keep_top_n=None,
                          use_signatures=None,
                          ignore_signatures=None):
    """

    Run the whole pipeline of decomposing DNV spectra for the samples specified in sample_names.

    :param DNVspectrumDict: dictionary containing DNV spectra as values and sample names as keys (dictionary)
    :param sample_names: The list of sample names analysed. (list of str)
    :param unique_only: If True, only unique mutations are used to construct the original spectrum.
    :param signatures_file: The path to the csv file containing the signature matrix. (str)
    :param equal_initial_proportions: If True, initial weights are initialized to be equal for all reference signature. Otherwise, they are initialized so the signatures with large cosine similarity with the original spectrum get larger weights. (default: False) (bool)
    :param tol: The maximal difference between the proportions for convergence to be True. (default: 0.0001) (float)
    :param max_iter: The maximum number of iterations (default: 1000) (int)
    :param filter_percent: Filter signatures, that contribute less than filter_percent of the mutations in the sample. (default: 0) (float)
    :param filter_count: Filter signatures, that contribute less than filter_count number of mutations in the sample. (default: 0) (int)
    :param keep_top_n: Only keep those signatures that contribute to the mixture with the top keep_top_n number of mutation. (default: None) (int)
    :param use_signatures: Use a specific subset of all signatures. A list of signature names. (default: None) (list of str)
    :param ignore_signatures: Exclude a specific subset of all signatures from the analysis. A list of signature names. (default: None) (list of str)

    :returns: The final proportions of all signatures in the mixture. (Filtered out or not used signatures appear with a proportion of zero.) (numpy.array)

    """

    if sample_names == None:
        sample_names = sorted(list(DNVspectrumDict.keys()))

    if signatures_file == None:
        signatures_file = HOME + '/alexandrovSignatures/sigProfiler_DBS_signatures.csv'

    sig_matrix, sig_names = __read_signatures_files(signatures_file, sig_type='DNV')

    sig_names_filtered = sig_names
    if use_signatures != None:
        sig_names_filtered = sorted(use_signatures)
    if ignore_signatures != None:
        sig_names_filtered = sorted(list(set(sig_names_filtered) - set(ignore_signatures)))

    theta_matrix = []
    for s in sample_names:
        theta = __decompose_spectrum(DNVspectrumDict[s], sig_matrix, sig_names,
                                     equal_initial_proportions=equal_initial_proportions,
                                     tol=tol,
                                     max_iter=max_iter,
                                     filter_percent=filter_percent,
                                     filter_count=filter_count,
                                     keep_top_n=keep_top_n,
                                     use_signatures=sig_names_filtered)
        theta_matrix.append(theta)

    return plot.__plot_spectrum_decomposition(sample_names, theta_matrix, sig_names_filtered, spectrum_type='DNV')


def decompose_SNV_spectra(SNVspectrumDict,
                          sample_names=None,
                          signatures_file=None,
                          equal_initial_proportions=False,
                          tol=0.0001,
                          max_iter=1000,
                          filter_percent=0,
                          filter_count=0,
                          keep_top_n=None,
                          use_signatures=None,
                          ignore_signatures=None):
    """

    Run the whole pipeline of decomposing SNV spectra for the samples specified in sample_names.

    :param SNVspectrumDict: dictionary containing SNV spectra as values and sample names as keys (dictionary)
    :param sample_names: The list of sample names analysed. (list of str)
    :param unique_only: If True, only unique mutations are used to construct the original spectrum.
    :param signatures_file: The path to the csv file containing the signature matrix. (str)
    :param equal_initial_proportions: If True, initial weights are initialized to be equal for all reference signature. Otherwise, they are initialized so the signatures with large cosine similarity with the original spectrum get larger weights. (default: False) (bool)
    :param tol: The maximal difference between the proportions for convergence to be True. (default: 0.0001) (float)
    :param max_iter: The maximum number of iterations (default: 1000) (int)
    :param filter_percent: Filter signatures, that contribute less than filter_percent of the mutations in the sample. (default: 0) (float)
    :param filter_count: Filter signatures, that contribute less than filter_count number of mutations in the sample. (default: 0) (int)
    :param keep_top_n: Only keep those signatures that contribute to the mixture with the top keep_top_n number of mutation. (default: None) (int)
    :param use_signatures: Use a specific subset of all signatures. A list of signature names. (default: None) (list of str)
    :param ignore_signatures: Exclude a specific subset of all signatures from the analysis. A list of signature names. (default: None) (list of str)

    :returns: The final proportions of all signatures in the mixture. (Filtered out or not used signatures appear with a proportion of zero.) (numpy.array)

    """

    if sample_names == None:
        sample_names = sorted(list(SNVspectrumDict.keys()))

    if signatures_file == None:
        signatures_file = HOME + '/alexandrovSignatures/sigProfiler_SBS_signatures_2018_03_28.csv'

    sig_matrix, sig_names = __read_signatures_files(signatures_file, sig_type='SNV')

    sig_names_filtered = sig_names
    if use_signatures != None:
        sig_names_filtered = sorted(use_signatures)
    if ignore_signatures != None:
        sig_names_filtered = sorted(list(set(sig_names_filtered) - set(ignore_signatures)))

    theta_matrix = []
    for s in sample_names:
        theta = __decompose_spectrum(SNVspectrumDict[s], sig_matrix, sig_names,
                                     equal_initial_proportions=equal_initial_proportions,
                                     tol=tol,
                                     max_iter=max_iter,
                                     filter_percent=filter_percent,
                                     filter_count=filter_count,
                                     keep_top_n=keep_top_n,
                                     use_signatures=sig_names_filtered)
        theta_matrix.append(theta)

    return plot.__plot_spectrum_decomposition(sample_names, theta_matrix, sig_names_filtered, spectrum_type='SNV')



try:
    from isomut2py import process
    from isomut2py import postprocess
    from isomut2py import io
    from isomut2py import plot
    from isomut2py import SAMTOOLS_MAX_DEPTH

    from datetime import datetime as datetime
    import time as time
    import os as os
    import subprocess as subprocess
except ImportError:
    print('ImportError in isomut2py.mutationcalling, MutationCaller object cannot be defined.')


class MutationCaller:
    '''

    The MutationCaller class is designed to keep all parameter values, directories and filepaths in one place that are needed for the mutation
    detection and postprocessing of a single or multiple sample(s).

    - List of basic parameters:
        - ref_fasta: The path to the fasta file of the reference genome. (str)

        - output_dir: The path to a directory that can be used for temporary files and output files. The user must have permission to write the directory. (str)

        - input_dir: The path to the directory, where the bam file(s) of the sample(s) is/are located. (str)

        - bam_filename: A list of the name(s) of the bam file(s) of the sample(s). (Without path, eg. ["sample_1.bam", "sample_2.bam", "sample_3.bam", ...].) (list of str)

        - samtools_fullpath: The path to samtools on the computer. (default: "samtools") (str)

    - Other parameters with default values:
        - n_min_block: The approximate number of blocks to partition the analysed genome to for parallel computing. The actual number might be slightly larger that this. (default: 200) (int)

        - n_conc_blocks: The number of blocks to process at the same time. (default: 4) (int)

        - chromosomes: The list of chromosomes to analyse. (default: all chromosomes included included in the reference genome specified in ref_fasta) (list of str)

        - base_quality_limit: The base quality limit used by samtools in order to decide if a base should be included in the pileup file. (default: 30) (int)

        - samtools_flags: The samtools flags to be used for pileup file generation. (default: " -B -d 1000 ") (str)

        - unique_mutations_only: If True, only those mutations are sought that are unique to specific samples. Setting it to False greatly increases computation time, but allows for the detection of shared mutations and thus the analysis of phylogenetic connections between samples. If True, print_shared_by_all is ignored. (default: True) (boolean)

        - print_shared_by_all: If False, mutations that are present in all analysed samples are not printed to the output files. This decreases both memory usage and computation time. (default: False) (boolean)

        - min_sample_freq: The minimum frequency of the mutated base at a given position in the mutated sample(s). (default: 0.21) (float)

        - min_other_ref_freq: The minimum frequency of the reference base at a given position in the non-mutated sample(s). (default: 0.95) (float)

        - cov_limit: The minimum coverage at a given position in the mutated sample(s). (default: 5) (float)

        - min_gap_dist_snv: Minimum genomic distance from an identified SNV for a position to be considered as a potential mutation. (default: 0) (int)

        - min_gap_dist_indel: Minimum genomic distance from an identified indel for a position to be considered as a potential mutation. (default: 20) (int)

        - use_local_realignment: By default, mutation detection is run only once, with the samtools mpileup command with option -B. This turns off the probabilistic realignment of reads while creating the temporary pileup file, which might result in false positives due to alignment error. To filter out these mutations, setting the above parameter to True runs the whole mutation detection pipeline again for possibly mutated positions without the -B option as well, and only those mutations are kept that are still present with the probabilistic realignment turned on. Setting use_local_realignment = True increases runtime. (default: False) (boolean)

        - ploidy_info_filepath: Path to the file containing ploidy information (see below for details) of the samples. (default: None) (str)

        - constant_ploidy: The default ploidy to be used in regions not contained in ploidy_info_filepath. (default: 2) (int)

        - bedfile: Path to a bedfile containing a list of genomic regions, if mutations are only sought in a limited part of the genome. (default: None) (str)

        - control_samples: list of the bam filenames of the control samples. Control samples are not expected to have any unique mutations, thus they can be used to optimize the results. (default: None) (array-like)

        - FPs_per_genome: the maximum number of false positive mutations allowed in a control sample (default: None) (int)

        - mutations: a pandas DataFrame of mutations found in the sample set (default: None) (pandas.DataFrame)

        - chrom_length: the length of chromosomes (default: not set) (array-like)

        - genome_length: the total length of the genome (default: not set) (int)

        - IDspectra: a dictionary containing the indel spectra of all the samples in the dataset (default: not set) (dict with keys=bam_filename and values=array-like)

        - DNVspectra: a dictionary containing the dinucleotide variation spectra of all the samples in the dataset (default: not set) (dict with keys=bam_filename and values=array-like)

        - SNVspectra: a dictionary containing the single nucleotide variation spectra of all the samples in the dataset (default: not set) (dict with keys=bam_filename and values=array-like)

    '''

    def __init__(self, **kwargs):
        """

        Initializes the MutationCaller object by setting attributes to the ones defined in kwargs.

        :param kwargs: keyword arguments

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

    def __check_params(self, list_of_params):
        """

        Checks and updates parameters values listed in list_of_param.

        :param list_of_params: parameter values to check (array-like)

        """
        def get_default_chroms_with_len(ref_genome, chrom_list):
            """

            Gets the length of chromosomes defined in chrom_list or if chrom_list is empty, gets the default list of chromosomes from a specified
            reference genome fasta file (set in ref_genome) with their length.

            :param ref_genome: The path to the reference genome fasta file.
            :param chrom_list: A predefined list of chromosomes to be included.

            :return: (chrom, lens):
                - chrom: a list of detected chromosomes (list of str)
                - len: a list of their lengths (list of int)

            """
            newchroms, newlens = [], []
            with open(ref_genome + '.fai') as f_h:
                for line in f_h:
                    chrom, leng = line.split('\t')[0], line.split('\t')[1]
                    if (chrom in chrom_list or len(chrom_list) == 0):
                        newchroms.append(chrom)
                        newlens.append(int(leng))
            if len(chrom_list) > 0:
                return chrom_list, [newlens[newchroms.index(c)] for c in chrom_list]
            else:
                return sorted(newchroms), [newlens[newchroms.index(c)] for c in sorted(newchroms)]

        defaults = dict()
        defaults['unique_mutations_only'] = True
        defaults['min_sample_freq'] = 0.21
        defaults['min_other_ref_freq'] = 0.95
        defaults['cov_limit'] = 5
        defaults['min_gap_dist_snv'] = 0
        defaults['min_gap_dist_indel'] = 20
        defaults['use_local_realignment'] = False
        defaults['ploidy_info_filepath'] = 'no_ploidy_info'
        defaults['constant_ploidy'] = 2
        defaults['print_shared_by_all'] = False
        defaults['n_min_block'] = 200
        defaults['n_conc_blocks'] = 4
        defaults['chromosomes'] = None
        defaults['base_quality_limit'] = 30
        defaults['bedfile'] = None
        defaults['samtools_flags'] = ' -B -d ' + str(SAMTOOLS_MAX_DEPTH) + ' '
        defaults['samtools_fullpath'] = 'samtools'
        defaults['control_samples'] = None
        defaults['FPs_per_genome'] = None
        defaults['mutations'] = None

        for p in list_of_params:
            if p in ['output_dir', 'input_dir', 'bam_filename']:
                if not hasattr(self, p):
                    raise ValueError('Error: Value of "' + p + '" is not set, cannot proceed.')
                elif p == 'input_dir' or p == 'bam_filename':
                    if not os.path.isdir(self.input_dir):
                        raise ValueError(
                            'Error: Input directory "' + self.input_dir + '" does not exist, cannot proceed.')
                    for bm in self.bam_filename:
                        if not os.path.isfile(self.input_dir + '/' + bm):
                            raise ValueError(
                                'Error: Bam file "' + self.input_dir + '/' + bm + '" does not exist, '
                                                                                  'cannot proceed.')
            elif p == 'samtools_fullpath':
                if not hasattr(self, p):
                    setattr(self, p, 'samtools')
                s = subprocess.getoutput(self.samtools_fullpath).strip().split(' ')[0] == 'Program:'
                if not s:
                    error_msg = 'Error: samtools not found, cannot proceed.'
                    error_msg += '\n'
                    error_msg += 'Specify path to samtools with the "samtools_fullpath" keyword argument.'
                    raise ValueError(error_msg)
            elif p == 'ref_fasta':
                if not hasattr(self, p):
                    raise ValueError('Error: Value of "' + p + '" is not set, cannot proceed.')
                if not os.path.isfile(self.ref_fasta):
                    raise ValueError(
                        'Error: Reference genome file "' + self.ref_fasta + '" does not exist, cannot proceed.')
                if not os.path.isfile(self.ref_fasta + '.fai'):
                    error_msg = 'Error: No faidx file found for reference genome file "' + self.ref_fasta + '", cannot proceed.'
                    error_msg += '\n'
                    error_msg += 'Use the samtools command: samtools faidx [ref.fasta]'
                    raise ValueError(error_msg)
            elif p == 'chromosomes':
                if not hasattr(self, 'chromosomes') or self.chromosomes is None or not hasattr(self, 'chrom_length'):
                    if not os.path.isfile(self.ref_fasta):
                        raise ValueError(
                            'Error: Reference genome file "' + self.ref_fasta + '" does not exist, cannot proceed.')
                    if not os.path.isfile(self.ref_fasta + '.fai'):
                        error_msg = 'Error: No faidx file found for reference genome file "' + self.ref_fasta + '", cannot proceed.'
                        error_msg += '\n'
                        error_msg += 'Use the samtools command: samtools faidx [ref.fasta]'
                        raise ValueError(error_msg)
                    if hasattr(self, 'chromosomes') and self.chromosomes is not None:
                        oldchrom = self.chromosomes
                    else:
                        oldchrom = []
                    c, l = get_default_chroms_with_len(self.ref_fasta, oldchrom)
                    setattr(self, 'chromosomes', c)
                    setattr(self, 'chrom_length', l)
                    setattr(self, 'genome_length', sum(l))
            elif p in ['IDspectra', 'DNVspectra', 'SNVspectra']:
                if not hasattr(self, p):
                    raise ValueError('Error: ' + p + ' dictionary not yet defined.')
            elif not hasattr(self, p):
                setattr(self, p, defaults[p])

            if p == 'min_sample_freq' and (self.min_sample_freq < 0 or self.min_sample_freq > 1):
                print('Value of "min_sample_freq" not in range [0,1], using default value of 0.21.')
                setattr(self, 'min_sample_freq', defaults['min_sample_freq'])
            if p == 'min_other_ref_freq' and (self.min_other_ref_freq < 0 or self.min_other_ref_freq > 1):
                print('Value of "min_other_ref_freq" not in range [0,1], using default value of 0.95.')
                setattr(self, 'min_other_ref_freq', defaults['min_other_ref_freq'])
            if p == 'ploidy_info_filepath' and not os.path.isfile(self.ploidy_info_filepath):
                print('File ' + self.ploidy_info_filepath + ' does not exist. Using constant default ploidy of 2.')
                setattr(self, 'ploidy_info_filepath', defaults['ploidy_info_filepath'])

    def run_isomut2_mutdet(self, **kwargs):
        """

        Runs IsoMut2 mutation detection pipeline on the MutationDetection object, using parameter values specified in the respective attributes of
        the object.

        :param kwargs: keyword arguments for MutationDetection object attributes

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['n_min_block',
                             'n_conc_blocks',
                             'chromosomes',
                             'ref_fasta',
                             'bam_filename',
                             'input_dir',
                             'output_dir',
                             'samtools_fullpath',
                             'samtools_flags',
                             'bedfile',
                             'min_sample_freq',
                             'min_other_ref_freq',
                             'cov_limit',
                             'base_quality_limit',
                             'min_gap_dist_snv',
                             'min_gap_dist_indel',
                             'constant_ploidy',
                             'ploidy_info_filepath',
                             'unique_mutations_only',
                             'print_shared_by_all',
                             'use_local_realignment'])

        if self.print_shared_by_all:
            msg = 'Warning: printing mutations shared by all samples might necessiate a large amount of disk space. ' \
                  'If your samples are largely different from the reference genome used and are generally similar to each other, ' \
                  'you should consider ' \
                  'skipping mutations shared by all of your samples. ' \
                  'To do this, set the attribute "print_shared_by_all" to False. ' \
                  '\n\n'
            print(msg)

        starting_time = datetime.now()
        level = 0

        print('\t' * level + time.strftime("%Y-%m-%d %H:%M:%S",
                                           time.localtime()) + ' - Mutation detection with IsoMut2\n')

        if (not self.use_local_realignment):  # single run only without local realignment
            process.single_run(mutcallObject=self, level=level)
        else:
            process.double_run(mutcallObject=self, level=level)

        finish_time = datetime.now()
        total_time = finish_time - starting_time
        total_time_h = int(total_time.seconds / 3600)
        total_time_m = int((total_time.seconds % 3600) / 60)
        total_time_s = (total_time.seconds % 3600) % 60
        print('\n' + time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()) + ' - IsoMut2 finished. (' + str(
            total_time.days) + ' day(s), ' + str(total_time_h) + ' hour(s), ' + str(total_time_m) + ' min(s), ' + str(
            total_time_s) + ' sec(s).)')

    def optimize_results(self, control_samples, FPs_per_genome, **kwargs):
        """

        Optimizes the list of detected mutations according to the list of control samples and desired level of false positives set by the user.
        Filtered results will be loaded to the mutations attribute of the MutationDetection object as a pandas.DataFrame. Optimized values for the
        score are stored in the attribute optimized_score_values.

        :param control_samples: List of sample names that should be used as control samples in the sense, that no unique mutations are expected in them. (The sample names listed here must match a subset of the sample names listed in the attribute bam_filename.) (list of str)
        :param FPs_per_genome: The total number of false positives tolerated in a control sample. (int)
        :param kwargs: possible keyword arguments besides MutationCaller attributes:

            - plot_roc_curve: If True, ROC curves will be plotted as a visual representation of the optimization process. (default: False) (boolean)
            - plot_tuning_curve: If True, tuning curves displaying the number of mutations found in different samples with different score filters will be plotted as a visual representation of the optimization process. (default: False) (boolean)

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        setattr(self, 'control_samples', control_samples)
        setattr(self, 'FPs_per_genome', FPs_per_genome)

        self.__check_params(['bam_filename', 'output_dir'])

        score_lim_dict, filtered_muts = postprocess.optimize_results(sample_names=self.bam_filename,
                                                                     control_samples=self.control_samples,
                                                                     FPs_per_genome=self.FPs_per_genome,
                                                                     plot_roc=kwargs.get('plot_roc', False),
                                                                     plot_tuning_curve=kwargs.get('plot_tuning_curve',
                                                                                                  False),
                                                                     output_dir=self.output_dir)

        self.optimized_score_values = score_lim_dict
        self.mutations = filtered_muts

    def load_mutations(self, filename=None):
        """

        Loads mutations from a file or a list of files into the mutations attribute.

        :param filename: The path to the file, where mutations are stored. A list of paths can be also supplied, in this case, all of them will be loaded to a single dataframe. The mutations attribute of the MutationDetection object will be set to the loaded dataframe. If None, the file [output_dir]/filtered_mutations.csv will be loaded. (default: None) (str)

        """

        self.__check_params(['output_dir'])

        df = io.load_mutations(output_dir=self.output_dir,
                               filename=filename)

        self.mutations = df

    def plot_mutation_counts(self,
                             unique_only=False,
                             return_string=False,
                             mutations_filename=None,
                             mutations_dataframe=None,
                             **kwargs):
        """

        Plots the number of mutations found in all the samples in different ploidy regions.

        :param unique_only: If True, only unique mutations are plotted for each sample. (default: False) (boolean)
        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)
        :param mutations_filename: The path to the file, where mutations are stored, if the mutations attribute of the object does not exist, its value will be set to the file defined here. (default: None) (str)
        :param mutations_dataframe: If the mutations are not to be loaded from a file, but are contained in a pandas.DataFrame, this can be supplied with setting the mutations_dataframe parameter. (default: None) (pandas.DataFrame)
        :param kwargs: possible keyword arguments besides MutationCaller attributes

            - control_samples: List of sample names that should be used as control samples in the sense, that no unique mutations are expected in them. (The sample names listed here must match a subset of the sample names listed in bam_filename.) (list of str)

        :returns: If the return_string value is True, a base64 encoded string of the image. Otherwise, a matplotlib figure.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['bam_filename', 'output_dir', 'control_samples'])

        return plot.plot_mutation_counts(sample_names=self.bam_filename,
                                         mutations_dataframe=mutations_dataframe,
                                         unique_only=unique_only,
                                         return_string=return_string,
                                         mutations_filename=mutations_filename,
                                         output_dir=self.output_dir,
                                         control_samples=self.control_samples)

    def plot_hierarchical_clustering(self,
                                     mutations_dataframe=None,
                                     return_string=False,
                                     mutations_filename=None,
                                     **kwargs):
        """

        Generates a heatmap based on the number of shared mutations found in all possible sample pairs.
        A dendrogram is also added that is the result of hierarchical clustering of the samples.

        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)
        :param mutations_filename: The path to the file, where mutations are stored, if the mutations attribute of the object does not exist, its value will be set to the file defined here. (default: None) (str)
        :param mutations_dataframe: If the mutations are not to be loaded from a file, but are contained in a pandas.DataFrame, this can be supplied with setting the mutations_dataframe parameter. (default: None) (pandas.DataFrame)
        :param kwargs: keyword arguments for MutationDetection object attributes

        :returns: If the return_string value is True, a base64 encoded string of the image. Otherwise, a matplotlib figure.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['bam_filename', 'output_dir'])

        return plot.plot_hierarchical_clustering(sample_names=self.bam_filename,
                                                 mutations_dataframe=mutations_dataframe,
                                                 mutations_filename=mutations_filename,
                                                 output_dir=self.output_dir,
                                                 return_string=return_string)

    def calculate_SNV_spectrum(self, unique_only=True, **kwargs):
        """

        Calculates the triplet spectrum from the dataframe of relevant mutations, using the fasta file of the reference genome. Results are stored
        in the SNVspectra attribute of the object.

        :param unique_only: If True, only unique mutations are plotted for each sample. (default: True) (boolean)
        :param kwargs: keyword arguments for MutationCaller attributes

            - mutations_dataframe: A pandas DataFrame where mutations are listed. (default: not set) (pandas.DataFrame)
            - sample_names: A list of sample names to plot results for, a subset of the list in "bam_filename" attribute. (default: not set) (list)
            - mutations_filaname: the path to the file where mutations are stored (default: not set) (str)

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['bam_filename', 'output_dir', 'chromosomes', 'ref_fasta', 'mutations'])

        SNVspectra = postprocess.calculate_SNV_spectrum(ref_fasta=self.ref_fasta,
                                                        mutations_dataframe=kwargs.get('mutations_dataframe',
                                                                                       self.mutations),
                                                        sample_names=kwargs.get('sample_names', self.bam_filename),
                                                        unique_only=unique_only,
                                                        chromosomes=self.chromosomes,
                                                        output_dir=self.output_dir,
                                                        mutations_filename=kwargs.get('mutations_filename', None))
        self.SNVspectra = SNVspectra

    def plot_SNV_spectrum(self, return_string=False, normalize_to_1=False, **kwargs):
        """

        Plots the triplet spectra for the samples in attribute bam_filename.

        :param normalize_to_1: If True, results are plotted as percentages, instead of counts. (default: False) (bool)
        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)

        :returns: If the return_string value is True, a list of base64 encoded strings of the images. Otherwise, a list of matplotlib figures.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['SNVspectra'])

        return plot.plot_SNV_spectrum(spectrumDict=self.SNVspectra,
                                      return_string=return_string,
                                      normalize_to_1=normalize_to_1)


    def calculate_and_plot_SNV_spectrum(self,
                                        unique_only=True,
                                        normalize_to_1=False,
                                        return_string=False,
                                        **kwargs):
        """

        Calculates and plots the SNV spectra of the samples listed in attribute bam_filename.

        :param unique_only: If True, only unique mutations are plotted for each sample. (default: True) (boolean)
        :param normalize_to_1: If True, results are plotted as percentages, instead of counts. (default: False) (bool)
        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)
        :param kwargs: keyword arguments for MutationCaller attributes

        :returns: If the return_string value is True, a list of base64 encoded strings of the images. Otherwise, a list of matplotlib figures.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.calculate_SNV_spectrum(unique_only=unique_only, **kwargs)
        return self.plot_SNV_spectrum(return_string=return_string, normalize_to_1=normalize_to_1, **kwargs)

    def calculate_indel_spectrum(self, unique_only=True, **kwargs):
        """

        Calculates the indel spectrum from the dataframe of relevant mutations, using the fasta file of the reference genome.
        Results are stored in the IDspectra attribute of the object.

        :param unique_only: If True, only unique mutations are plotted for each sample. (default: True) (boolean)
        :param kwargs: keyword arguments for MutationCaller attributes

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['bam_filename', 'output_dir', 'chromosomes', 'ref_fasta', 'mutations'])

        IDspectra = postprocess.calculate_indel_spectrum(ref_fasta=self.ref_fasta,
                                                         mutations_dataframe=kwargs.get('mutations_dataframe',
                                                                                        self.mutations),
                                                         sample_names=kwargs.get('sample_names', self.bam_filename),
                                                         unique_only=unique_only,
                                                         chromosomes=self.chromosomes,
                                                         output_dir=self.output_dir,
                                                         mutations_filename=kwargs.get('mutations_filename', None))
        self.IDspectra = IDspectra

    def plot_indel_spectrum(self, return_string=False, normalize_to_1=False, **kwargs):
        """

        Plots the indel spectra for the samples in attribute bam_filename.

        :param normalize_to_1: If True, results are plotted as percentages, instead of counts. (default: False) (bool)
        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)

        :returns: If the return_string value is True, a list of base64 encoded strings of the images. Otherwise, a list of matplotlib figures.

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['IDspectra'])

        return plot.plot_indel_spectrum(spectrumDict=self.IDspectra,
                                        return_string=return_string,
                                        normalize_to_1=normalize_to_1)

    def calculate_and_plot_indel_spectrum(self,
                                          unique_only=True,
                                          normalize_to_1=False,
                                          return_string=False,
                                          **kwargs):
        """

        Calculates and plots the indel spectra of the samples listed in attribute bam_filename.

        :param unique_only: If True, only unique mutations are plotted for each sample. (default: True) (boolean)
        :param normalize_to_1: If True, results are plotted as percentages, instead of counts. (default: False) (bool)
        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)
        :param kwargs: keyword arguments for MutationCaller attributes

        :returns: If the return_string value is True, a list of base64 encoded strings of the images. Otherwise, a list of matplotlib figures.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.calculate_indel_spectrum(unique_only=unique_only, **kwargs)
        return self.plot_indel_spectrum(return_string=return_string, normalize_to_1=normalize_to_1, **kwargs)

    def calculate_DNV_matrix(self, unique_only=True, **kwargs):
        """

        Calculates the dinucleotide variation spectrum in matrix format from the dataframe of relevant mutations, using the fasta file of the
        reference genome. Results are stored in the DNVmatrice attribute of the object.

        :param unique_only: If True, only unique mutations are plotted for each sample. (default: True) (boolean)
        :param kwargs: keyword arguments for MutationCaller attributes

            - mutations_dataframe: A pandas DataFrame where mutations are listed. (default: not set) (pandas.DataFrame)
            - sample_names: A list of sample names to plot results for, a subset of the list in "bam_filename" attribute. (default: not set) (list)
            - mutations_filaname: the path to the file where mutations are stored (default: not set) (str)

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['bam_filename', 'output_dir', 'chromosomes', 'ref_fasta', 'mutations'])

        DNVmatrice = postprocess.calculate_DNV_matrix(
            mutations_dataframe=kwargs.get('mutations_dataframe', self.mutations),
            sample_names=kwargs.get('sample_names', self.bam_filename),
            unique_only=unique_only,
            chromosomes=self.chromosomes,
            output_dir=self.output_dir,
            mutations_filename=kwargs.get('mutations_filename', None))

        self.DNVmatrice = DNVmatrice

    def plot_DNV_heatmap(self, return_string=False, **kwargs):
        """

        Plots the DNV spectra as a heatmap for the samples in attribute bam_filename.

        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)

        :returns: If the return_string value is True, a list of base64 encoded strings of the images. Otherwise, a list of matplotlib figures.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['DNVmatrice'])

        return plot.plot_DNV_heatmap(matrixDict=self.DNVmatrice,
                                     return_string=return_string)

    def calculate_and_plot_DNV_heatmap(self,
                                       unique_only=True,
                                       return_string=False,
                                       **kwargs):
        """

        Calculates and plots the DNV spectra as a heatmap of the samples listed in attribute bam_filename.

        :param unique_only: If True, only unique mutations are plotted for each sample. (default: True) (boolean)
        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)
        :param kwargs: keyword arguments for MutationCaller attributes

        :returns: If the return_string value is True, a list of base64 encoded strings of the images. Otherwise, a list of matplotlib figures.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.calculate_DNV_matrix(unique_only=unique_only, **kwargs)
        return self.plot_DNV_heatmap(return_string=return_string, **kwargs)

    def calculate_DNV_spectrum(self, unique_only=True, **kwargs):
        """

        Calculates the dinucleotide variation spectrum from the dataframe of relevant mutations, using the fasta file of the reference genome.
        Results are stored in the DNVmatrice attribute of the object.

        :param unique_only: If True, only unique mutations are plotted for each sample. (default: True) (boolean)
        :param kwargs: keyword arguments for MutationCaller attributes

            - mutations_dataframe: A pandas DataFrame where mutations are listed. (default: not set) (pandas.DataFrame)
            - sample_names: A list of sample names to plot results for, a subset of the list in "bam_filename" attribute. (default: not set) (list)
            - mutations_filaname: the path to the file where mutations are stored (default: not set) (str)

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['bam_filename', 'output_dir', 'chromosomes', 'ref_fasta', 'mutations'])

        DNVspectra = postprocess.calculate_DNV_spectrum(
            mutations_dataframe=kwargs.get('mutations_dataframe', self.mutations),
            sample_names=kwargs.get('sample_names', self.bam_filename),
            unique_only=unique_only,
            chromosomes=self.chromosomes,
            output_dir=self.output_dir,
            mutations_filename=kwargs.get('mutations_filename', None))

        self.DNVspectra = DNVspectra

    def plot_DNV_spectrum(self, return_string=False, normalize_to_1=False, **kwargs):
        """

        Plots the DNV spectra for the samples in attribute bam_filename.

        :param normalize_to_1: If True, results are plotted as percentages, instead of counts. (default: False) (bool)
        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)

        :returns: If the return_string value is True, a list of base64 encoded strings of the images. Otherwise, a list of matplotlib figures.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['DNVspectra'])

        return plot.plot_DNV_spectrum(spectrumDict=self.DNVspectra,
                                      return_string=return_string,
                                      normalize_to_1=normalize_to_1)

    def calculate_and_plot_DNV_spectrum(self,
                                        unique_only=True,
                                        return_string=False,
                                        normalize_to_1=False,
                                        **kwargs):
        """

        Calculates and plots the DNV spectra of the samples listed in attribute bam_filename.

        :param unique_only: If True, only unique mutations are plotted for each sample. (default: True) (boolean)
        :param normalize_to_1: If True, results are plotted as percentages, instead of counts. (default: False) (bool)
        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)
        :param kwargs: keyword arguments for MutationCaller attributes

        :returns: If the return_string value is True, a list of base64 encoded strings of the images. Otherwise, a list of matplotlib figures.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.calculate_DNV_spectrum(unique_only=unique_only, **kwargs)
        return self.plot_DNV_spectrum(return_string=return_string, normalize_to_1=normalize_to_1, **kwargs)

    def plot_rainfall(self,
                      unique_only=True,
                      return_string=False,
                      **kwargs):
        """

        Plots the rainfall plot of mutations found in the samples listed in the attribute bam_filename. Displaying rainfall plots is a good
        practice to detect mutational clusters throughout the genome. The horizontal axis is the genomic position of each mutation,
        while the vertical shows the genomic distance between the mutation and the previous one. Thus mutations that are clustered together appear
        close to each other horizontally and on the lower part of the plot vertically.

        :param unique_only: If True, only unique mutations are plotted for each sample. (default: True) (boolean)
        :param return_string: If True, only a temporary plot is generated and its base64 code is returned, that can be included in HTML files. (default: False) (bool)
        :param kwargs: keyword arguments for MutationCaller attributes

            - mutations_dataframe: A pandas DataFrame where mutations are listed. (default: not set) (pandas.DataFrame)
            - sample_names: A list of sample names to plot results for, a subset of the list in "bam_filename" attribute. (default: not set) (list)
            - mut_types: list of mutation types to display (default: not set, displaying SNVs, insertions and deletions) (a list containing any combination of these items: ['SNV', 'INS', 'DEL'])
            - plot_range: The genomic range to plot. (default: not set) (str, example: "chr9:2342-24124")

        :returns: If the return_string value is True, a list of base64 encoded strings of the images. Otherwise, a list of matplotlib figures.

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['bam_filename', 'chromosomes', 'mutations'])

        return plot.plot_rainfall(mutations_dataframe=kwargs.get('mutations_dataframe', self.mutations),
                                  chromosomes=self.chromosomes,
                                  chrom_length=self.chrom_length,
                                  sample_names=kwargs.get('sample_names', self.bam_filename),
                                  return_string=return_string,
                                  muttypes=kwargs.get('muttypes', ['SNV', 'INS', 'DEL']),
                                  unique_only=unique_only,
                                  plot_range = kwargs.get('plot_range', None))

    def check_pileup(self, chrom_list, from_pos_list, to_pos_list, print_original=True, savetofile=None,
                     **kwargs):
        """

        Loads pileup information for a list of genomic regions.

        :param chrom_list: List of chromosomes for the regions. (list of str)
        :param from_pos_list: List of starting positions for the regions. (list of int)
        :param to_pos_list: List of ending positions for the regions. (list of int)
        :param print_original: If True, prints the original string generated with samtools mpileup as well (default: True) (bool)
        :param savetofile: If savetofile is specified, results will be writted to that path. Otherwise, results are written to [output_dir]/checkPileup_tmp.csv (default: None) (str)

        :returns: df: The processed results of the pileup, containing information on all samples in a pandas.DataFrame.

        """

        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['ref_fasta', 'input_dir', 'bam_filename', 'output_dir', 'samtools_fullpath',
                             'base_quality_limit', 'samtools_flags'])

        return postprocess.check_pileup(chrom_list=chrom_list,
                                        from_pos_list=from_pos_list,
                                        to_pos_list=to_pos_list,
                                        ref_fasta=self.ref_fasta,
                                        input_dir=self.input_dir,
                                        bam_filename=self.bam_filename,
                                        output_dir=self.output_dir,
                                        samtools_fullpath=self.samtools_fullpath,
                                        base_quality_limit=self.base_quality_limit,
                                        samtools_flags=self.samtools_flags,
                                        print_original=print_original,
                                        filename=savetofile)

    def get_details_for_mutations(self,
                                  mutations_filename=None,
                                  **kwargs):
        """

        Get detailed results for the list of mutations contained in the mutations attribute of the object.

        :param mutations_filename: The path(s) to the file(s) where mutations are stored. (default: None) (list of str)
        :param kwargs: keyword arguments for MutationDetection object attributes

            - mutations_dataframe: A pandas DataFrame where mutations are listed. (default: not set) (pandas.DataFrame)

        :returns: df_joined: A dataframe containing the detailed results. (pandas.DataFrame)

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

        self.__check_params(['ref_fasta', 'input_dir', 'bam_filename', 'output_dir', 'samtools_fullpath',
                             'base_quality_limit', 'samtools_flags', 'mutations'])

        return postprocess.get_details_for_mutations(ref_fasta=self.ref_fasta,
                                                     input_dir=self.input_dir,
                                                     bam_filename=self.bam_filename,
                                                     mutations_dataframe=kwargs.get('mutations_dataframe',
                                                                                    self.mutations),
                                                     output_dir=self.output_dir,
                                                     mutations_filename=mutations_filename,
                                                     samtools_fullpath=self.samtools_fullpath,
                                                     base_quality_limit=self.base_quality_limit,
                                                     samtools_flags=self.samtools_flags)

    def decompose_indel_spectra(self,
                                sample_names=None,
                                unique_only=True,
                                signatures_file=None,
                                equal_initial_proportions=False,
                                tol=0.0001,
                                max_iter=1000,
                                filter_percent=0,
                                filter_count=0,
                                keep_top_n=None,
                                use_signatures=None,
                                ignore_signatures=None,
                                **kwargs):
        """

        Run the whole pipeline of decomposing indel spectra for the samples specified in sample_names.

        :param sample_names: The list of sample names analysed. (list of str)
        :param unique_only: If True, only unique mutations are used to construct the original spectrum.
        :param signatures_file: The path to the csv file containing the signature matrix. (str)
        :param equal_initial_proportions: If True, initial weights are initialized to be equal for all reference signature. Otherwise, they are initialized so the signatures with large cosine similarity with the original spectrum get larger weights. (default: False) (bool)
        :param tol: The maximal difference between the proportions for convergence to be True. (default: 0.0001) (float)
        :param max_iter: The maximum number of iterations (default: 1000) (int)
        :param filter_percent: Filter signatures, that contribute less than filter_percent of the mutations in the sample. (default: 0) (float)
        :param filter_count: Filter signatures, that contribute less than filter_count number of mutations in the sample. (default: 0) (int)
        :param keep_top_n: Only keep those signatures that contribute to the mixture with the top keep_top_n number of mutation. (default: None) (int)
        :param use_signatures: Use a specific subset of all signatures. A list of signature names. (default: None) (list of str)
        :param ignore_signatures: Exclude a specific subset of all signatures from the analysis. A list of signature names. (default: None) (list of str)
        :param kwargs: keyword arguments for MutationDetection object attributes

        :returns: The final proportions of all signatures in the mixture. (Filtered out or not used signatures appear with a proportion of zero.) (numpy.array)

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

        if not hasattr(self, 'IDspectra') or self.IDspectra.__class__ != dict:
            print('Indel spectra have not been calculated yet, starting with this step.')
            self.calculate_indel_spectrum(unique_only=unique_only, **kwargs)

        return postprocess.decompose_indel_spectra(IDspectrumDict=self.IDspectra,
                                                   sample_names=sample_names,
                                                   signatures_file=signatures_file,
                                                   equal_initial_proportions=equal_initial_proportions,
                                                   tol=tol,
                                                   max_iter=max_iter,
                                                   filter_percent=filter_percent,
                                                   filter_count=filter_count,
                                                   keep_top_n=keep_top_n,
                                                   use_signatures=use_signatures,
                                                   ignore_signatures=ignore_signatures)

    def decompose_DNV_spectra(self,
                              sample_names=None,
                              unique_only=True,
                              signatures_file=None,
                              equal_initial_proportions=False,
                              tol=0.0001,
                              max_iter=1000,
                              filter_percent=0,
                              filter_count=0,
                              keep_top_n=None,
                              use_signatures=None,
                              ignore_signatures=None,
                              **kwargs):
        """

        Run the whole pipeline of decomposing DNV spectra for the samples specified in sample_names.

        :param sample_names: The list of sample names analysed. (list of str)
        :param unique_only: If True, only unique mutations are used to construct the original spectrum.
        :param signatures_file: The path to the csv file containing the signature matrix. (str)
        :param equal_initial_proportions: If True, initial weights are initialized to be equal for all reference signature. Otherwise, they are initialized so the signatures with large cosine similarity with the original spectrum get larger weights. (default: False) (bool)
        :param tol: The maximal difference between the proportions for convergence to be True. (default: 0.0001) (float)
        :param max_iter: The maximum number of iterations (default: 1000) (int)
        :param filter_percent: Filter signatures, that contribute less than filter_percent of the mutations in the sample. (default: 0) (float)
        :param filter_count: Filter signatures, that contribute less than filter_count number of mutations in the sample. (default: 0) (int)
        :param keep_top_n: Only keep those signatures that contribute to the mixture with the top keep_top_n number of mutation. (default: None) (int)
        :param use_signatures: Use a specific subset of all signatures. A list of signature names. (default: None) (list of str)
        :param ignore_signatures: Exclude a specific subset of all signatures from the analysis. A list of signature  names. (default: None) (list of str)
        :param kwargs: keyword arguments for MutationDetection object attributes

        :returns: The final proportions of all signatures in the mixture. (Filtered out or not used signatures appear with a proportion of zero.) (numpy.array)

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

        if not hasattr(self, 'DNVspectra') or self.DNVspectra.__class__ != dict:
            print('DNV spectra have not been calculated yet, starting with this step.')
            self.calculate_DNV_spectrum(unique_only=unique_only, **kwargs)

        return postprocess.decompose_DNV_spectra(DNVspectrumDict=self.DNVspectra,
                                                 sample_names=sample_names,
                                                 signatures_file=signatures_file,
                                                 equal_initial_proportions=equal_initial_proportions,
                                                 tol=tol,
                                                 max_iter=max_iter,
                                                 filter_percent=filter_percent,
                                                 filter_count=filter_count,
                                                 keep_top_n=keep_top_n,
                                                 use_signatures=use_signatures,
                                                 ignore_signatures=ignore_signatures)

    def decompose_SNV_spectra(self,
                              sample_names=None,
                              unique_only=True,
                              signatures_file=None,
                              equal_initial_proportions=False,
                              tol=0.0001,
                              max_iter=1000,
                              filter_percent=0,
                              filter_count=0,
                              keep_top_n=None,
                              use_signatures=None,
                              ignore_signatures=None,
                              **kwargs):
        """

        Run the whole pipeline of decomposing SNV spectra for the samples specified in sample_names.

        :param sample_names: The list of sample names analysed. (list of str)
        :param unique_only: If True, only unique mutations are used to construct the original spectrum.
        :param signatures_file: The path to the csv file containing the signature matrix. (str)
        :param equal_initial_proportions: If True, initial weights are initialized to be equal for all reference signature. Otherwise, they are initialized so the signatures with large cosine similarity with the original spectrum get larger weights. (default: False) (bool)
        :param tol: The maximal difference between the proportions for convergence to be True. (default: 0.0001) (float)
        :param max_iter: The maximum number of iterations (default: 1000) (int)
        :param filter_percent: Filter signatures, that contribute less than filter_percent of the mutations in the sample. (default: 0) (float)
        :param filter_count: Filter signatures, that contribute less than filter_count number of mutations in the sample. (default: 0) (int)
        :param keep_top_n: Only keep those signatures that contribute to the mixture with the top keep_top_n number of mutation. (default: None) (int)
        :param use_signatures: Use a specific subset of all signatures. A list of signature names. (default: None) (list of str)
        :param ignore_signatures: Exclude a specific subset of all signatures from the analysis. A list of signature names. (default: None) (list of str)
        :param kwargs: keyword arguments for MutationDetection object attributes

        :returns: The final proportions of all signatures in the mixture. (Filtered out or not used signatures appear with a proportion of zero.) (numpy.array)

        """
        for k, v in kwargs.items():
            setattr(self, k, v)

        if not hasattr(self, 'SNVspectra') or self.SNVspectra.__class__ != dict:
            print('SNV spectra have not been calculated yet, starting with this step.')
            self.calculate_SNV_spectrum(unique_only=unique_only, **kwargs)

        return postprocess.decompose_SNV_spectra(SNVspectrumDict=self.SNVspectra,
                                                 sample_names=sample_names,
                                                 signatures_file=signatures_file,
                                                 equal_initial_proportions=equal_initial_proportions,
                                                 tol=tol,
                                                 max_iter=max_iter,
                                                 filter_percent=filter_percent,
                                                 filter_count=filter_count,
                                                 keep_top_n=keep_top_n,
                                                 use_signatures=use_signatures,
                                                 ignore_signatures=ignore_signatures)



from setuptools import setup

setup(
    name='isomut2py',
    version='2.0.5',
    description='A mutation detection software from NGS data with multiple filtering and plotting features',
    url='http://github.com/pipekorsi/isomut2',
    author='Orsolya Pipek',
    author_email='pipeko@caesar.elte.hu',
    license='MIT',
    packages=['isomut2py'],
    package_data={'isomut2py': ['alexandrovSignatures/*.csv', 'C/*']},
    install_requires=['matplotlib>=2.0.0',
                      'pandas>=0.20.3',
                      'Theano>=1.0.4',
                      'pymc3==3.1',
                      'setuptools>=33.1.1',
                      'seaborn>=0.8',
                      'scipy==0.19.1',
                      'numpy==1.12.1',
                      'biopython']
)

##### THIS DOES NOT SEEM TO WORK...
##### (when packaged with wheel, it runs, but not at the installation, and this is also mentioned as a problem on SO)

# import atexit
# import os
# import sys
# from setuptools import setup
# from setuptools.command.install import install
# import subprocess
#
#
# class CustomInstall(install):
#     def run(self):
#         def _post_install():
#             def find_module_path():
#                 for p in sys.path:
#                     if os.path.isdir(p) and 'isomut2py' in os.listdir(p):
#                         return os.path.join(p, 'isomut2py')
#
#             install_path = find_module_path()
#
#             # Add your post install code here
#
#             for __fname in os.listdir(install_path + '/C'):
#                 if __fname.split('.')[-1] != 'h' and __fname.split('.')[-1] != 'c':
#                     subprocess.call('rm ' + install_path + '/C/' + __fname, shell=True)
#
#             print('Compiling C scripts...')
#             cwd = os.getcwd()
#             __compiled = []
#             os.chdir(install_path + '/C')
#             __cmd = 'gcc -c -O3 isomut2_lib.c fisher.c  -W -Wall'
#             __compiled.append(subprocess.call(__cmd, shell=True))
#             __cmd = 'gcc -O3 -o isomut2 isomut2.c isomut2_lib.o fisher.o -lm -W -Wall'
#             __compiled.append(subprocess.call(__cmd, shell=True))
#             __cmd = 'gcc -O3 -o PEprep PEprep.c isomut2_lib.o fisher.o -lm -W -Wall'
#             __compiled.append(subprocess.call(__cmd, shell=True))
#             __cmd = 'gcc -O3 -o checkPileup checkPileup.c isomut2_lib.o fisher.o -lm -W -Wall'
#             __compiled.append(subprocess.call(__cmd, shell=True))
#             os.chdir(cwd)
#
#             if 1 in __compiled:
#                 print('WARNING: C scripts were not compiled correctly, some functions will not work.')
#                 print('Try compiling manually to ' + install_path + '/C')
#             else:
#                 print('Done.')
#
#         atexit.register(_post_install)
#         install.run(self)
#
#
# setup(
#     cmdclass={'install': CustomInstall},
#     name='isomut2py',
#     version='2.0.2',
#     description='A mutation detection software from NGS data with multiple filtering and plotting features',
#     url='http://github.com/pipekorsi/isomut2',
#     author='Orsolya Pipek',
#     author_email='pipeko@caesar.elte.hu',
#     license='MIT',
#     packages=['isomut2py'],
#     package_data={'isomut2py': ['alexandrovSignatures/*.csv', 'C/*']},
#     install_requires=['matplotlib>=2.0.0',
#                       'pandas>=0.20.3',
#                       'Theano>=0.9.0',
#                       'pymc3>=3.1',
#                       'setuptools>=33.1.1',
#                       'seaborn>=0.8',
#                       'scipy>=0.19.1',
#                       'numpy>=1.12.1',
#                       'biopython']
# )



try:
    from isomut2py import ploidyestimation

    import pandas as __pd
    import numpy as __np
    import os as __os
    import subprocess as __subprocess
except ImportError:
    print('ImportError in isomut2py.format, generating bed files will not work.')


def get_bed_format_for_sample(chromosomes, chrom_length, output_dir, bam_filename=None, ownbed_filepath=None):
    """

    Creates bed file of constant ploidies for a given sample from a file of positional ploidy data.
    If the ownbed_filepath attribute of the PloidyEstimation object is set, saves the bedfile to the path specified there. Otherwise,
    saves it to the output_dir with the "_ploidy.bed" suffix. Also sets the bed_dataframe attribute to the pandas.Dataframe containing the bed file.

    :param chromosomes: list of chromosomes (array-like)
    :param chrom_length: list of chromosome lengths in bp (array-like)
    :param output_dir: path to the directory where the PE_fullchrom_* files are located. (str)
    :param bam_filename: filename of the BAM file of the sample (default: None) (str)
    :param ownbed_filepath: path to the bed file where results should be saved (default: None) (str)

    :returns: (ownbed_filepath, df)

        - ownbed_filepath: path the the bed file where results are saved
        - df: the bed file in a pandas.DataFrame

    """

    def get_ploidy_ranges_for_chrom(outputdir, chrom, chr_list, before_list, after_list, pl_list, loh_list,
                                    chrom_len_dict):
        """

        Gets ranges of constant ploidies from a file of positional ploidy data.
        At breakpoints, the rounded average of the two bordering positions are taken as the breakpoint position.

        :param outputdir: The path to the directory, where the PE_fullchrom_[chrom].txt file is located. (str)
        :param chrom: The name of the chromosome to be analysed. (str)
        :param chr_list: The list of chromosomes to append current information to.
        :param before_list: The list of starting positions to append current information to.
        :param after_list: The list of ending positions to append current information to.
        :param pl_list: The list of ploidies to append current information to.
        :param loh_list: The list of LOHs to append current information to.
        :param chrom_len_dict: A dictionary containing {chromosome: chromosome length} pairs.

        :return: (chr_list, before_list, after_list, pl_list, loh_list)

        """
        df = __pd.read_csv(outputdir + '/PE_fullchrom_' + chrom + '.txt', sep='\t').sort_values(by='pos')
        p = __np.array(list(df['pos']))
        pl = __np.array(list(df['ploidy']))
        loh = __np.array(list(df['LOH']))
        pl_loh = __np.array([str(pl_c) + ',' + str(loh_c) for pl_c, loh_c in zip(pl, loh)])
        pl_loh_change = __np.where(pl_loh[:-1] != pl_loh[1:])[0]

        before_pos = 0
        after_pos = 0
        for i in range(len(pl_loh_change)):
            after_idx = pl_loh_change[i]
            after_pos = int(round(__np.mean([p[after_idx], p[after_idx + 1]])))
            chr_list.append(chrom)
            before_list.append(before_pos)
            after_list.append(after_pos)
            pl_list.append(pl[after_idx])
            loh_list.append(loh[after_idx])
            before_pos = after_pos + 1
        chr_list.append(chrom)
        before_list.append(before_pos)
        after_list.append(chrom_len_dict[chrom])
        pl_list.append(pl[-1])
        loh_list.append(loh[-1])

        return chr_list, before_list, after_list, pl_list, loh_list

    chrom_len_dict = {c: l for c, l in zip(chromosomes, chrom_length)}
    chr_list = []
    before_list = []
    after_list = []
    pl_list = []
    loh_list = []
    for c in chromosomes:
        if not __os.path.isfile(output_dir + '/PE_fullchrom_' + c + '.txt'):
            raise ValueError(
                'File ' + output_dir + '/PE_fullchrom_' + c + '.txt is not yet created, call "run_ploidy_estimation" '
                                                              'first.')
        chr_list, before_list, after_list, pl_list, loh_list = get_ploidy_ranges_for_chrom(output_dir, c,
                                                                                           chr_list, before_list,
                                                                                           after_list, pl_list,
                                                                                           loh_list, chrom_len_dict)
    df = __pd.DataFrame()
    df['chrom'] = chr_list
    df['chromStart'] = before_list
    df['chromEnd'] = after_list
    df['ploidy'] = pl_list
    df['LOH'] = loh_list

    if ownbed_filepath != None:
        df.to_csv(ownbed_filepath, index=False)
    elif bam_filename != None:
        df.to_csv(output_dir + '/' + bam_filename.split('.bam')[0] + '_ploidy.bed', index=False)
        ownbed_filepath = output_dir + '/' + bam_filename.split('.bam')[0] + '_ploidy.bed'
    else:
        df.to_csv(output_dir + '/' + 'ploidy.bed', index=False)
        ownbed_filepath = output_dir + '/' + 'ploidy.bed'

    return ownbed_filepath, df


def generate_ploidy_info_file(filename=None, sample_names=None, bed_filepaths=None, ploidy_estimation_objects=None,
                              sample_groups=None, group_bed_filepaths=None):
    """

    Generate ploidy info file for mutation detection in samples with nondefault ploidies. Make sure to supply one of the following arguments:
        - ploidy_estimation_objects
        - sample_names AND bed_filepaths
        - sample_groups AND group_bed_filepaths

    :param filename: The desired path to the generated ploidy info file. If None, ploidy information is saved to ploidy_info_file.txt in the current directory. (default: None) (str)
    :param sample_names: List of bam filenames for the samples, must be supplied together with bed_filepaths. (default: None) (list of str)
    :param bed_filepaths: Must be supplied together with sample_names. The list of filepaths to each bed file describing the given sample in sample_names. (default: None) (list of str)
    :param ploidy_estimation_objects: List of PloidyEstimation objects for each sample. (default: None) (list of isomut2py.PloidyEstimation)
    :param sample_groups: List of lists of str. Each list in sample_groups must contain the name of bam files in that group. Must be supplied together with group_bed_filepaths. (default: None) (list of list of str, example: [['sample1.bam', 'sample2.bam', 'sample3.bam'], ['sample4.bam', 'sample5.bam'], ['sample6.bam']])
    :param group_bed_filepaths: List of filepaths to the bed files describing each sample group in sample_groups. Must be supplied together with sample_groups. (default: None) (list of str, example: ['bedfile_of_samples123.txt', 'bedfile_of_samples45.txt' 'bedfile_of_samples6.txt'])

    """
    if filename == None:
        filename = 'ploidy_info_file.txt'
        print('Argument "filename" not set, saving ploidy info file to ploidy_info_file.txt')

    if (ploidy_estimation_objects == None and
            not (sample_names != None and bed_filepaths != None)
            and not (sample_groups != None and group_bed_filepaths != None)):
        raise TypeError(
            'Either "ploidy_estimation_objects" OR "sample_names" together with "bed_filepaths" OR "sample_groups" '
            'together with "group_bed_filepaths" must be defined.')

    __subprocess.call('rm ' + filename, shell=True)

    with open(filename, "a") as f:
        f.write('#file_path\tsample_names_list\n')
        if ploidy_estimation_objects != None:
            if ploidy_estimation_objects.__class__ != list:
                raise TypeError(
                    'Argument "ploidy_estimation_objects" must be a list.')
            for PE in ploidy_estimation_objects:
                if not isinstance(PE, ploidyestimation.PloidyEstimator):
                    raise TypeError('All elements of ploidy_estimation_objects must be PloidyEstimator objects.')
                if not hasattr(PE, "ownbed_filepath"):
                    raise ValueError(
                        'All PloidyEstimator objects must have valid "ownbed_filepath" arguments. Try running '
                        '"run_ploidy_estimation()" or set "ownbed_filepath" manually.')
                if not __os.path.isfile(PE.ownbed_filepath):
                    raise ValueError(
                        'All PloidyEstimator objects must have valid "ownbed_filepath" arguments. Try running '
                        '"run_ploidy_estimation()" or set "ownbed_filepath" manually.')
                else:
                    s = __subprocess.check_output('head -1 ' + PE.ownbed_filepath, shell=True)
                    sep = ','
                    if (len(s.decode('utf-8').strip().split('\t')) > 1):
                        sep = '\t'
                    elif (len(s.decode('utf-8').strip().split(';')) > 1):
                        sep = ';'
                    df = __pd.read_csv(PE.ownbed_filepath, sep=sep)
                    if "chrom" not in list(df.columns) or "chromStart" not in list(
                            df.columns) or "chromEnd" not in list(df.columns) or "ploidy" not in list(df.columns):
                        raise ValueError(
                            "File " + PE.ownbed_filepath + "does not have one of the following columns: chrom, "
                                                           "chromStart, chromEnd, ploidy.")
                    else:
                        df = df[['chrom', 'chromStart', 'chromEnd', 'ploidy']]
                        df.to_csv(PE.ownbed_filepath + '_im2format', sep='\t', index=False)
                        f.write(PE.ownbed_filepath + '_im2format\t' + PE.bam_filename + '\n')
        elif sample_names != None and bed_filepaths != None:
            if sample_names.__class__ != list:
                raise TypeError(
                    'Argument "sample_names" must be a list.')
            elif bed_filepaths.__class__ != list:
                raise TypeError(
                    'Argument "bed_filepaths" must be a list.')
            elif len(sample_names) != len(bed_filepaths):
                raise ValueError('Arguments "sample_names" and "bed_filepaths" must have the same length.')
            else:
                for sn, bfp in zip(sample_names, bed_filepaths):
                    if (not __os.path.isfile(bfp)):
                        raise ValueError('File ' + bfp + ' does not exist.')
                    else:
                        s = __subprocess.check_output('head -1 ' + bfp, shell=True)
                        sep = ','
                        if len(s.decode('utf-8').strip().split('\t')) > 1:
                            sep = '\t'
                        elif len(s.decode('utf-8').strip().split(';')) > 1:
                            sep = ';'
                        df = __pd.read_csv(bfp, sep=sep)
                        if ("chrom" not in list(df.columns) or "chromStart" not in list(
                                df.columns) or "chromEnd" not in list(df.columns) or "ploidy" not in list(df.columns)):
                            raise ValueError(
                                "File " + bfp + "does not have one of the following columns: chrom, chromStart, "
                                                "chromEnd, ploidy.")
                        else:
                            df = df[['chrom', 'chromStart', 'chromEnd', 'ploidy']]
                            df.to_csv(bfp + '_im2format', sep='\t', index=False)
                            f.write(bfp + '_im2format\t' + sn + '\n')
        elif sample_groups != None and group_bed_filepaths != None:
            if sample_groups.__class__ != list:
                raise TypeError(
                    'Argument "sample_groups" must be a list of lists.')
            for sg in sample_groups:
                if sg.__class__ != list:
                    raise TypeError(
                        'Argument "sample_groups" must be a list of lists.')
            if group_bed_filepaths.__class__ != list:
                raise TypeError(
                    'Argument "group_bed_filepaths" must be a list.')
            elif len(sample_groups) != len(group_bed_filepaths):
                raise ValueError('Arguments "sample_groups" and "group_bed_filepaths" must have the same length.')
            else:
                for sg, gbfp in zip(sample_groups, group_bed_filepaths):
                    if not __os.path.isfile(gbfp):
                        raise ValueError('File ' + gbfp + ' does not exist.')
                    else:
                        s = __subprocess.check_output('head -1 ' + gbfp, shell=True)
                        sep = ','
                        if len(s.decode('utf-8').strip().split('\t')) > 1:
                            sep = '\t'
                        elif len(s.decode('utf-8').strip().split(';')) > 1:
                            sep = ';'
                        df = __pd.read_csv(gbfp, sep=sep)
                        if ("chrom" not in list(df.columns) or "chromStart" not in list(
                                df.columns) or "chromEnd" not in list(df.columns) or "ploidy" not in list(df.columns)):
                            raise ValueError(
                                "File " + gbfp + "does not have one of the following columns: chrom, chromStart, "
                                                 "chromEnd, ploidy.")
                        else:
                            df = df[['chrom', 'chromStart', 'chromEnd', 'ploidy']]
                            df.to_csv(gbfp + '_im2format', sep='\t', index=False)
                            f.write(gbfp + '_im2format\t' + ', '.join(sg) + '\n')
